# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""VISCERAL dataset for 3D organ classification (only lungs and bladders).

Loaded samples are not full scans but 16x16x16 volumes of organs.
Database reference:
"""

import os
import pathlib

import torchio as tio

from ....data.datamodule import CachingDataModule
from ....data.split import make_split
from ....data.typing import RawDataLoader as _BaseRawDataLoader
from ....data.typing import Sample
from ....utils.rc import load_rc

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the VISCERAL dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self) -> None:
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Load a single volume sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the volume to be loaded and an integer, representing
            the sample label.

        Returns
        -------
            The sample representation.
        """
        clamp = tio.Clamp(out_min=-1000, out_max=2000)
        rescale = tio.RescaleIntensity(percentiles=(0.5, 99.5))
        preprocess = tio.Compose([clamp, rescale])
        image = tio.ScalarImage(self.datadir / sample[0])
        image = preprocess(image)
        tensor = image.data
        return tensor, dict(label=sample[1], name=sample[0])

    def label(self, sample: tuple[str, int]) -> int:
        """Load a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample label.

        Returns
        -------
        int
            The integer label associated with the sample.
        """

        return sample[1]


class DataModule(CachingDataModule):
    """VISCERAL DataModule for 3D organ binary classification.

    Data specifications:

    * Raw data input (on disk):

      * NIfTI volumes
      * resolution: 16x16x16

    * Output image:

      * Transforms:

        * Load raw NIfTI with :py:mod:`torchio`
        * Clamp and Rescale intensity
        * Convert to torch tensor

      * Final specifications

        * 32-bit floats, cubes 16x16x16 pixels
        * Labels: 0 (bladder), 1 (lung)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        super().__init__(
            make_split(__package__, split_filename),
            raw_data_loader=RawDataLoader(),
            database_name=__package__.split(".")[-1],
            split_name=pathlib.Path(split_filename).stem,
        )
