# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Indian database for TB detection (a.k.a. Dataset A/Dataset B).

Database reference: [INDIAN-2013]_
"""

import pathlib

from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import make_split

from ....config.data.shenzhen.datamodule import ClassificationRawDataLoader

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class DataModule(CachingDataModule):
    """Indian database for TB detection (a.k.a. Dataset A/Dataset B).

    The Indian collection database has been established to foster research in
    computer-aided diagnosis of pulmonary diseases with a special focus on
    pulmonary tuberculosis (TB).  This database is also known as the "Database
    A/Database B" database.

    * Database reference: [INDIAN-2013]_
    * Original images PNG, 8-bit grayscale, 1024 x 1024 pixels
    * Split reference: [INDIAN-2013]_ with 20% of train set for the validation
      set

    Data specifications:

    * Raw data input (on disk):

      * PNG RGB 8-bit depth images with "inverted" grayscale scale
      * Variable width and height

    * Output image:

      * Transforms:

        * Load raw PNG with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

      * Final specifications:

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square, with varying resolutions, depending on the input raw image
        * Labels: 0 (healthy), 1 (active tuberculosis)

    Parameters
    ----------
    split_filename
      Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=ClassificationRawDataLoader(
                config_variable=CONFIGURATION_KEY_DATADIR
            ),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
