# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of Montgomery and Shenzhen databases."""

import pathlib

from mednet.libs.common.data.datamodule import ConcatDataModule
from mednet.libs.common.data.split import make_split

from ..montgomery.datamodule import (
    ClassificationRawDataLoader as MontgomeryLoader,
)
from ..shenzhen.datamodule import ClassificationRawDataLoader as ShenzhenLoader


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of Montgomery and Shenzhen datasets.

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        montgomery_loader = MontgomeryLoader()
        montgomery_package = MontgomeryLoader.__module__.rsplit(".", 1)[0]
        montgomery_split = make_split(montgomery_package, split_filename)
        shenzhen_loader = ShenzhenLoader()
        shenzhen_package = ShenzhenLoader.__module__.rsplit(".", 1)[0]
        shenzhen_split = make_split(shenzhen_package, split_filename)

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                ],
            },
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
