# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of Montgomery, Shenzhen, Indian, and PadChest
datasets.
"""

import pathlib

from mednet.libs.common.data.datamodule import ConcatDataModule
from mednet.libs.common.data.split import make_split

from ..indian.datamodule import CONFIGURATION_KEY_DATADIR as INDIAN_KEY_DATADIR
from ..indian.datamodule import ClassificationRawDataLoader as IndianLoader
from ..indian.datamodule import DataModule as IndianDataModule
from ..montgomery.datamodule import (
    ClassificationRawDataLoader as MontgomeryLoader,
)
from ..padchest.datamodule import ClassificationRawDataLoader as PadchestLoader
from ..shenzhen.datamodule import ClassificationRawDataLoader as ShenzhenLoader


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of Montgomery, Shenzhen, Indian, and
    PadChest datasets.

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    padchest_split_filename
        Name of the .json file from padchest containing the split to load.
    """

    def __init__(self, split_filename: str, padchest_split_filename: str):
        assert __package__ is not None
        montgomery_loader = MontgomeryLoader()
        montgomery_package = MontgomeryLoader.__module__.rsplit(".", 1)[0]
        montgomery_split = make_split(montgomery_package, split_filename)
        shenzhen_loader = ShenzhenLoader()
        shenzhen_package = ShenzhenLoader.__module__.rsplit(".", 1)[0]
        shenzhen_split = make_split(shenzhen_package, split_filename)
        indian_loader = IndianLoader(INDIAN_KEY_DATADIR)
        indian_package = IndianDataModule.__module__.rsplit(".", 1)[0]
        indian_split = make_split(indian_package, split_filename)
        padchest_loader = PadchestLoader()
        padchest_package = PadchestLoader.__module__.rsplit(".", 1)[0]
        padchest_split = make_split(padchest_package, padchest_split_filename)

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                    (indian_split["train"], indian_loader),
                    (padchest_split["train"], padchest_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                    (indian_split["validation"], indian_loader),
                    # there is no validation set on padchest...
                    # (padchest_split["validation"], padchest_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                    (indian_split["test"], indian_loader),
                    (padchest_split["test"], padchest_loader),
                ],
            },
            database_name=__package__.rsplit(".", 1)[1],
            split_name=(
                pathlib.Path(split_filename).stem
                + "+"
                + pathlib.Path(padchest_split_filename).stem
            ),
        )
