# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""NIH CXR14 (relabeled) DataModule for computer-aided diagnosis.

Database reference: [NIH-CXR14-2017]_
"""

import os
import pathlib

import PIL.Image
from mednet.libs.classification.data.typing import (
    ClassificationRawDataLoader as _ClassificationRawDataLoader,
)
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.utils.rc import load_rc
from torchvision.transforms.functional import to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database.
"""

CONFIGURATION_KEY_IDIAP_FILESTRUCTURE = (
    (__name__.rsplit(".", 2)[-2]) + ".idiap_folder_structure"
)
"""Key to search for in the configuration file indicating if the loader should
use standard or idiap-based file organisation structure.

It causes the internal loader to search for files in a slightly
different folder structure, that was adapted to Idiap's requirements
(number of files per folder to be less than 10k).
"""


class ClassificationRawDataLoader(_ClassificationRawDataLoader):
    """A specialized raw-data-loader for the NIH CXR-14 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    idiap_file_organisation: bool
    """If should use the Idiap's filesystem organisation when looking up data.

    This variable will be ``True``, if the user has set the configuration
    parameter ``nih_cxr14.idiap_file_organisation`` in the global configuration
    file.  It will cause internal loader to search for files in a slightly
    different folder structure, that was adapted to Idiap's requirements
    (number of files per folder to be less than 10k).
    """

    def __init__(self):
        rc = load_rc()
        self.datadir = pathlib.Path(
            rc.get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir)),
        )
        self.idiap_file_organisation = rc.get(
            CONFIGURATION_KEY_IDIAP_FILESTRUCTURE,
            False,
        )

    def sample(self, sample: tuple[str, list[int]]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        file_path = pathlib.Path(sample[0])  # default
        if self.idiap_file_organisation:
            # for folder lookup efficiency, data is split into subfolders
            # each original file is on the subfolder `f[:5]/f`, where f
            # is the original file basename
            file_path = pathlib.Path(
                file_path.parent / file_path.name[:5] / file_path.name
            )

        # N.B.: some NIH CXR-14 images are encoded as color PNGs with an alpha
        # channel.  Most, are grayscale PNGs
        image = PIL.Image.open(self.datadir / file_path)
        image = image.convert("L")  # required for some images
        tensor = to_tensor(image)

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(target=sample[1], name=sample[0])  # type: ignore[arg-type]

    def target(self, sample: tuple[str, list[int]]) -> list[int]:
        """Load a single image sample target from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample target.

        Returns
        -------
        list[int]
            The integer targets associated with the sample.
        """

        return sample[1]


class DataModule(CachingDataModule):
    """NIH CXR14 (relabeled) DataModule for computer-aided diagnosis.

    This dataset was extracted from the clinical PACS database at the National
    Institutes of Health Clinical Center (USA) and represents 60% of all their
    radiographs. It contains labels for 14 common radiological signs in this
    order: cardiomegaly, emphysema, effusion, hernia, infiltration, mass,
    nodule, atelectasis, pneumothorax, pleural thickening, pneumonia, fibrosis,
    edema and consolidation. This is the relabeled version created in the
    CheXNeXt study.

    * Reference: [NIH-CXR14-2017]_
    * Raw data input (on disk):

      * PNG RGB 8-bit depth images
      * Resolution: 1024 x 1024 pixels

    * Labels: [CHEXNEXT-2018]_
    * Split reference: [CHEXNEXT-2018]_
    * Output image:

        * Transforms:

            * Load raw PNG with :py:mod:`PIL`
            * Convert to torch tensor

        * Final specifications:

            * RGB, encoded as a 3-plane tensor, 32-bit floats, square
              (1024x1024 px)
            * Labels in order:
              * cardiomegaly
              * emphysema
              * effusion
              * hernia
              * infiltration
              * mass
              * nodule
              * atelectasis
              * pneumothorax
              * pleural thickening
              * pneumonia
              * fibrosis
              * edema
              * consolidation

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=ClassificationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
