# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of NIH-CXR-14 and PadChest databases."""

import pathlib

from mednet.libs.common.data.datamodule import ConcatDataModule
from mednet.libs.common.data.split import make_split

from ..nih_cxr14.datamodule import ClassificationRawDataLoader as CXR14Loader
from ..padchest.datamodule import ClassificationRawDataLoader as PadchestLoader


class DataModule(ConcatDataModule):
    """Aggregated dataset composed of NIH CXR14 relabeld and PadChest
    (normalized) datasets.

    Parameters
    ----------
    cxr14_split_filename
        Name of the .json file from crx14 containing the split to load.
    padchest_split_filename
        Name of the .json file from padchest containing the split to load.
    """

    def __init__(self, cxr14_split_filename: str, padchest_split_filename):
        assert __package__ is not None
        cxr14_loader = CXR14Loader()
        cxr14_package = CXR14Loader.__module__.rsplit(".", 1)[0]
        cxr14_split = make_split(cxr14_package, cxr14_split_filename)
        padchest_loader = PadchestLoader()
        padchest_package = PadchestLoader.__module__.rsplit(".", 1)[0]
        padchest_split = make_split(padchest_package, padchest_split_filename)

        super().__init__(
            splits={
                "train": [
                    (cxr14_split["train"], cxr14_loader),
                    (padchest_split["train"], padchest_loader),
                ],
                "validation": [
                    (cxr14_split["validation"], cxr14_loader),
                    (padchest_split["validation"], padchest_loader),
                ],
                "test": [
                    (cxr14_split["test"], cxr14_loader),
                    # there is no test set on padchest
                    # (padchest_split["test"], padchest_loader),
                ],
            },
            database_name=__package__.rsplit(".", 1)[1],
            split_name=(
                pathlib.Path(cxr14_split_filename).stem
                + "+"
                + pathlib.Path(padchest_split_filename).stem
            ),
        )
