# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Shenzhen DataModule for computer-aided diagnosis.

Database reference: [MONTGOMERY-SHENZHEN-2014]_
"""

import os
import pathlib

import PIL.Image
from mednet.libs.classification.data.typing import (
    ClassificationRawDataLoader as _ClassificationRawDataLoader,
)
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.image_utils import remove_black_borders
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.utils.rc import load_rc
from torchvision.transforms.functional import to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class ClassificationRawDataLoader(_ClassificationRawDataLoader):
    """A specialized raw-data-loader for the Shenzhen dataset.

    Parameters
    ----------
    config_variable
        Key to search for in the configuration file for the root directory of
        this database.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    # config_variable: required so this loader can be used for the Indian
    # database as well.
    def __init__(self, config_variable: str = CONFIGURATION_KEY_DATADIR):
        self.datadir = pathlib.Path(
            load_rc().get(config_variable, os.path.realpath(os.curdir)),
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # N.B.: Image.convert("L") is required to normalize grayscale back to
        # normal (instead of inverted).
        image = PIL.Image.open(self.datadir / sample[0]).convert("L")
        image, _ = remove_black_borders(image)
        tensor = to_tensor(image)

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(target=sample[1], name=sample[0])  # type: ignore[arg-type]

    def target(self, sample: tuple[str, int]) -> int:
        """Load a single image sample target from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
        int
            The integer target associated with the sample.
        """

        return sample[1]


class DataModule(CachingDataModule):
    """Shenzhen DataModule for computer-aided diagnosis.

    The standard digital image database for Tuberculosis was created by the
    National Library of Medicine, Maryland, USA in collaboration with Shenzhen
    No.3 People’s Hospital, Guangdong Medical College, Shenzhen, China. The
    Chest X-rays are from out-patient clinics, and were captured as part of the
    daily routine using Philips DR Digital Diagnose systems.

    * Database reference: [MONTGOMERY-SHENZHEN-2014]_

    Data specifications:

    * Raw data input (on disk):

      * PNG 8-bit RGB images (grayscale, but encoded as RGB images with
        "inverted" grayscale scale requiring special treatment).
      * Variable width and height of 3000 x 3000 pixels or less

    * Output image:

      * Transforms:

        * Load raw PNG with :py:mod:`PIL`
        * Remove black borders
        * Torch center cropping to get square image

      * Final specifications:

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square with varying resolutions, depending on the input image
        * Labels: 0 (healthy), 1 (active tuberculosis)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=ClassificationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
