# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import pathlib

import PIL.Image
from mednet.libs.classification.data.typing import (
    ClassificationRawDataLoader as _ClassificationRawDataLoader,
)
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.image_utils import remove_black_borders
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.utils.rc import load_rc
from torchvision.transforms.functional import to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class ClassificationRawDataLoader(_ClassificationRawDataLoader):
    """A specialized raw-data-loader for the Shenzen dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # images from TBPOC are encoded as grayscale JPEGs, no need to
        # call convert("L") here.
        image = PIL.Image.open(self.datadir / sample[0])
        image, _ = remove_black_borders(image)
        tensor = to_tensor(image)

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(target=sample[1], name=sample[0])  # type: ignore[arg-type]

    def target(self, sample: tuple[str, int]) -> int:
        """Load a single image sample target from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The integer target associated with the sample
        """

        return sample[1]


class DataModule(CachingDataModule):
    """TB-POC dataset for computer-aided diagnosis.

    * Database reference: [TB-POC-2018]_
    * Original resolution (height x width or width x height): 2048 x 2500 pixels
      or 2500 x 2048 pixels

    Data specifications:

    * Raw data input (on disk):

      * JPEG 8-bit Grayscale images
      * resolution: fixed to one of the cases above

    * Output image:

      * Transforms:

        * Load raw grayscale jpeg with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

      * Final specifications:

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square with varying resolutions (2048 x 2048 being the maximum),
          but also depending on black borders' sizes on the input image.
        * Labels: 0 (healthy), 1 (active tuberculosis)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=ClassificationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
