# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines functionality for the evaluation of predictions."""

import contextlib
import itertools
import json
import logging
import typing
from collections.abc import Iterable, Iterator

import credible.curves
import matplotlib.figure
import numpy
import numpy.typing
import sklearn.metrics
import tabulate
from matplotlib import pyplot as plt

from ..models.typing import BinaryPrediction

logger = logging.getLogger("mednet")


def eer_threshold(predictions: Iterable[BinaryPrediction]) -> float:
    """Calculate the (approximate) threshold leading to the equal error rate.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.typing.BinaryPrediction`'s.

    Returns
    -------
    float
        The EER threshold value.
    """

    from scipy.interpolate import interp1d
    from scipy.optimize import brentq

    y_scores = [k[2] for k in predictions]
    y_labels = [k[1] for k in predictions]

    fpr, tpr, thresholds = sklearn.metrics.roc_curve(y_labels, y_scores)

    eer = brentq(lambda x: 1.0 - x - interp1d(fpr, tpr)(x), 0.0, 1.0)
    return float(interp1d(fpr, thresholds)(eer))


def _get_centered_maxf1(
    f1_scores: numpy.typing.NDArray,
    thresholds: numpy.typing.NDArray,
) -> tuple[float, float]:
    """Return the centered max F1 score threshold when multiple thresholds give
    the same max F1 score.

    Parameters
    ----------
    f1_scores
        1D array of f1 scores.
    thresholds
        1D array of thresholds.

    Returns
    -------
    tuple(float, float)
        A tuple with the maximum F1-score and the "centered" threshold.
    """

    maxf1 = f1_scores.max()
    maxf1_indices = numpy.where(f1_scores == maxf1)[0]

    # If multiple thresholds give the same max F1 score
    if len(maxf1_indices) > 1:
        mean_maxf1_index = int(round(numpy.mean(maxf1_indices)))
    else:
        mean_maxf1_index = maxf1_indices[0]

    return maxf1, thresholds[mean_maxf1_index]


def maxf1_threshold(predictions: Iterable[BinaryPrediction]) -> float:
    """Calculate the threshold leading to the maximum F1-score on a precision-
    recall curve.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.typing.BinaryPrediction`'s.

    Returns
    -------
    float
        The threshold value leading to the maximum F1-score on the provided set
        of predictions.
    """

    y_scores = [k[2] for k in predictions]
    y_labels = [k[1] for k in predictions]

    precision, recall, thresholds = sklearn.metrics.precision_recall_curve(
        y_labels,
        y_scores,
    )

    numerator = 2 * recall * precision
    denom = recall + precision
    f1_scores = numpy.divide(
        numerator,
        denom,
        out=numpy.zeros_like(denom),
        where=(denom != 0),
    )

    _, maxf1_threshold = _get_centered_maxf1(f1_scores, thresholds)
    return maxf1_threshold


def score_plot(
    histograms: dict[str, dict[str, numpy.typing.NDArray]],
    title: str,
    threshold: float | None,
) -> matplotlib.figure.Figure:
    """Plot the normalized score distributions for all systems.

    Parameters
    ----------
    histograms
        A dictionary containing all histograms that should be inserted into the
        plot.  Each histogram should itself be setup as another dictionary
        containing the keys ``hist`` and ``bin_edges`` as returned by
        :py:func:`numpy.histogram`.
    title
        Title of the plot.
    threshold
        Shows where the threshold is in the figure.  If set to ``None``, then
        does not show the threshold line.

    Returns
    -------
    matplotlib.figure.Figure
        A single (matplotlib) plot containing the score distribution, ready to
        be saved to disk or displayed.
    """

    from matplotlib.ticker import MaxNLocator

    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)
    ax = typing.cast(plt.Axes, ax)  # gets editor to behave

    # Here, we configure the "style" of our plot
    ax.set_xlim((0, 1))
    ax.set_title(title)
    ax.set_xlabel("Score")
    ax.set_ylabel("Count")

    # Only show ticks on the left and bottom spines
    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.get_yaxis().set_major_locator(MaxNLocator(integer=True))

    # Setup the grid
    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)
    ax.get_xaxis().grid(False)

    max_hist = 0
    for name in histograms.keys():
        hist = histograms[name]["hist"]
        bin_edges = histograms[name]["bin_edges"]
        width = 0.7 * (bin_edges[1] - bin_edges[0])
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        ax.bar(center, hist, align="center", width=width, label=name, alpha=0.7)
        max_hist = max(max_hist, hist.max())

    # Detach axes from the plot
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015 * max_hist))

    if threshold is not None:
        # Adds threshold line (dotted red)
        ax.axvline(
            threshold,  # type: ignore
            color="red",
            lw=2,
            alpha=0.75,
            ls="dotted",
            label="threshold",
        )

    # Adds a nice legend
    ax.legend(
        fancybox=True,
        framealpha=0.7,
    )

    # Makes sure the figure occupies most of the possible space
    fig.tight_layout()

    return fig


def run_binary(
    name: str,
    predictions: Iterable[BinaryPrediction],
    binning: str | int,
    threshold_a_priori: float | None = None,
) -> dict[str, typing.Any]:
    """Run inference and calculates measures for binary classification.

    Parameters
    ----------
    name
        The name of subset to load.
    predictions
        A list of predictions to consider for measurement.
    binning
        The binning algorithm to use for computing the bin widths and
        distribution for histograms.  Choose from algorithms supported by
        :py:func:`numpy.histogram`.
    threshold_a_priori
        A threshold to use, evaluated *a priori*, if must report single values.
        If this value is not provided, an *a posteriori* threshold is calculated
        on the input scores.  This is a biased estimator.

    Returns
    -------
    dict[str, typing.Any]
        A tuple containing the following entries:

        * summary: A dictionary containing the performance summary on the
          specified threshold, general performance curves (under the key
          ``curves``), and score histograms (under the key
          ``score-histograms``).
    """

    y_scores = numpy.array([k[2] for k in predictions])  # likelihoods
    y_labels = numpy.array([k[1] for k in predictions])  # integers
    neg_label = y_labels.min()
    pos_label = y_labels.max()

    use_threshold = threshold_a_priori
    if use_threshold is None:
        use_threshold = maxf1_threshold(predictions)
        logger.warning(
            f"User did not pass an *a priori* threshold for the evaluation "
            f"of split `{name}`.  Using threshold a posteriori (biased) with value "
            f"`{use_threshold:.4f}`",
        )

    y_predictions = numpy.where(y_scores >= use_threshold, pos_label, neg_label)

    # point measures on threshold
    summary = dict(
        num_samples=len(y_labels),
        threshold=use_threshold,
        threshold_a_posteriori=(threshold_a_priori is None),
        precision=sklearn.metrics.precision_score(
            y_labels,
            y_predictions,
            pos_label=pos_label,
        ),
        recall=sklearn.metrics.recall_score(
            y_labels,
            y_predictions,
            pos_label=pos_label,
        ),
        f1_score=sklearn.metrics.f1_score(
            y_labels,
            y_predictions,
            pos_label=pos_label,
        ),
        average_precision_score=sklearn.metrics.average_precision_score(
            y_labels,
            y_scores,
            pos_label=pos_label,
        ),
        specificity=sklearn.metrics.recall_score(
            y_labels,
            y_predictions,
            pos_label=neg_label,
        ),
        auc_score=sklearn.metrics.roc_auc_score(
            y_labels,
            y_scores,
        ),
        accuracy=sklearn.metrics.accuracy_score(y_labels, y_predictions),
    )

    # curves: ROC and precision recall
    summary["curves"] = dict(
        roc=dict(
            zip(
                ("fpr", "tpr", "thresholds"),
                sklearn.metrics.roc_curve(
                    y_labels,
                    y_scores,
                    pos_label=pos_label,
                ),
            ),
        ),
        precision_recall=dict(
            zip(
                ("precision", "recall", "thresholds"),
                sklearn.metrics.precision_recall_curve(
                    y_labels,
                    y_scores,
                    pos_label=pos_label,
                ),
            ),
        ),
    )

    # score histograms
    # what works: <integer>, doane*, scott, stone, rice*, sturges*, sqrt
    # what does not work: auto, fd
    summary["score-histograms"] = dict(
        positives=dict(
            zip(
                ("hist", "bin_edges"),
                numpy.histogram(
                    y_scores[y_labels == pos_label],
                    bins=binning,
                    range=(0, 1),
                ),
            ),
        ),
        negatives=dict(
            zip(
                ("hist", "bin_edges"),
                numpy.histogram(
                    y_scores[y_labels == neg_label],
                    bins=binning,
                    range=(0, 1),
                ),
            ),
        ),
    )

    return summary


def tabulate_results(
    data: typing.Mapping[str, typing.Mapping[str, typing.Any]],
    fmt: str,
) -> str:
    """Tabulate summaries from multiple splits.

    This function can properly tabulate the various summaries produced for all
    the splits in a prediction database.

    Parameters
    ----------
    data
        An iterable over all summary data collected.
    fmt
        One of the formats supported by `python-tabulate
        <https://pypi.org/project/tabulate/>`_.

    Returns
    -------
    str
        A string containing the tabulated information.
    """

    example = next(iter(data.values()))
    headers = list(example.keys())
    table = [[k[h] for h in headers] for k in data.values()]

    # add subset names
    headers = ["subset"] + headers
    table = [[name] + k for name, k in zip(data.keys(), table)]

    return tabulate.tabulate(table, headers, tablefmt=fmt, floatfmt=".3f")


def aggregate_roc(
    data: typing.Mapping[str, typing.Any],
    title: str = "ROC",
) -> matplotlib.figure.Figure:
    """Aggregate ROC curves from multiple splits.

    This function produces a single ROC plot for multiple curves generated per
    split.

    Parameters
    ----------
    data
        A dictionary mapping split names to ROC curve data produced by
        :py:func:sklearn.metrics.roc_curve`.
    title
        The title of the plot.

    Returns
    -------
    matplotlib.figure.Figure
        A figure, containing the aggregated ROC plot.
    """

    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)

    # Names and bounds
    ax.set_xlabel("1 - specificity")
    ax.set_ylabel("Sensitivity")
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.0])
    ax.set_title(title)

    # we should see some of ax 1 ax
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015))

    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)

    plt.tight_layout()

    lines = ["-", "--", "-.", ":"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]
    colorcycler = itertools.cycle(colors)
    linecycler = itertools.cycle(lines)

    legend = []

    for name, elements in data.items():
        # plots roc curve
        _auc = sklearn.metrics.auc(elements["fpr"], elements["tpr"])
        label = f"{name} (AUC={_auc:.2f})"
        color = next(colorcycler)
        style = next(linecycler)

        (line,) = ax.plot(
            elements["fpr"],
            elements["tpr"],
            color=color,
            linestyle=style,
        )
        legend.append((line, label))

    if len(legend) > 1:
        ax.legend(
            [k[0] for k in legend],
            [k[1] for k in legend],
            loc="lower right",
            fancybox=True,
            framealpha=0.7,
        )

    return fig


@contextlib.contextmanager
def _precision_recall_canvas() -> (
    Iterator[tuple[matplotlib.figure.Figure, matplotlib.figure.Axes]]
):
    """Generate a canvas to draw precision-recall curves.

    Works like a context manager, yielding a figure and an axes set in which
    the precision-recall curves should be added to.  The figure already
    contains F1-ISO lines and is preset to a 0-1 square region.  Once the
    context is finished, ``fig.tight_layout()`` is called.

    Yields
    ------
    figure
        The figure that should be finally returned to the user.
    axes
        An axis set where to precision-recall plots should be added to.
    """

    fig, axes1 = plt.subplots(1)
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(axes1, matplotlib.figure.Axes)

    # Names and bounds
    axes1.set_xlabel("Recall")
    axes1.set_ylabel("Precision")
    axes1.set_xlim([0.0, 1.0])
    axes1.set_ylim([0.0, 1.0])

    axes1.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)
    axes2 = axes1.twinx()

    # Annotates plot with F1-score iso-lines
    f_scores = numpy.linspace(0.1, 0.9, num=9)
    tick_locs = []
    tick_labels = []
    for f_score in f_scores:
        x = numpy.linspace(0.01, 1)
        y = f_score * x / (2 * x - f_score)
        plt.plot(x[y >= 0], y[y >= 0], color="green", alpha=0.1)
        tick_locs.append(y[-1])
        tick_labels.append(f"{f_score:.1f}")
    axes2.tick_params(axis="y", which="both", pad=0, right=False, left=False)
    axes2.set_ylabel("iso-F", color="green", alpha=0.3)
    axes2.set_ylim([0.0, 1.0])
    axes2.yaxis.set_label_coords(1.015, 0.97)
    axes2.set_yticks(tick_locs)  # notice these are invisible
    for k in axes2.set_yticklabels(tick_labels):
        k.set_color("green")
        k.set_alpha(0.3)
        k.set_size(8)

    # we should see some of axes 1 axes
    axes1.spines["right"].set_visible(False)
    axes1.spines["top"].set_visible(False)
    axes1.spines["left"].set_position(("data", -0.015))
    axes1.spines["bottom"].set_position(("data", -0.015))

    # we shouldn't see any of axes 2 axes
    axes2.spines["right"].set_visible(False)
    axes2.spines["top"].set_visible(False)
    axes2.spines["left"].set_visible(False)
    axes2.spines["bottom"].set_visible(False)

    # yield execution, lets user draw precision-recall plots, and the legend
    # before tighteneing the layout
    yield fig, axes1

    plt.tight_layout()


def aggregate_pr(
    data: typing.Mapping[str, typing.Any],
    title: str = "Precision-Recall Curve",
) -> matplotlib.figure.Figure:
    """Aggregate PR curves from multiple splits.

    This function produces a single Precision-Recall plot for multiple curves
    generated per split. The plot will be annotated with F1-score iso-lines (in
    which the F1-score maintains the same value).

    Parameters
    ----------
    data
        A dictionary mapping split names to Precision-Recall curve data produced by
        :py:func:sklearn.metrics.precision_recall_curve`.
    title
        The title of the plot.

    Returns
    -------
    matplotlib.figure.Figure
        A figure, containing the aggregated PR plot.
    """

    lines = ["-", "--", "-.", ":"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]
    colorcycler = itertools.cycle(colors)
    linecycler = itertools.cycle(lines)

    with _precision_recall_canvas() as (fig, axes):
        axes.set_title(title)
        legend = []

        for name, elements in data.items():
            _ap = credible.curves.average_metric(
                (elements["precision"], elements["recall"]),
            )
            label = f"{name} (AP={_ap:.2f})"
            color = next(colorcycler)
            style = next(linecycler)

            (line,) = axes.plot(
                elements["recall"],
                elements["precision"],
                color=color,
                linestyle=style,
            )
            legend.append((line, label))

        if len(legend) > 1:
            axes.legend(
                [k[0] for k in legend],
                [k[1] for k in legend],
                loc="lower left",
                fancybox=True,
                framealpha=0.7,
            )

    return fig


class NumpyJSONEncoder(json.JSONEncoder):
    """Extends the standard JSON encoder to support Numpy arrays."""

    def default(self, o: typing.Any) -> typing.Any:
        """If input object is a ndarray it will be converted into a list.

        Parameters
        ----------
        o
            Input object to be JSON serialized.

        Returns
        -------
            A serializable representation of object ``o``.
        """

        if isinstance(o, numpy.ndarray):
            try:
                retval = o.tolist()
            except TypeError:
                pass
            else:
                return retval

        # Let the base class default method raise the TypeError
        return super().default(o)
