# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

import lightning.pytorch
import torch.utils.data
from mednet.libs.common.engine.device import DeviceManager

from ..models.typing import (
    BinaryPrediction,
    BinaryPredictionSplit,
    MultiClassPrediction,
    MultiClassPredictionSplit,
)

logger = logging.getLogger("mednet")


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
) -> (
    list[BinaryPrediction]
    | list[MultiClassPrediction]
    | list[list[BinaryPrediction]]
    | list[list[MultiClassPrediction]]
    | BinaryPredictionSplit
    | MultiClassPredictionSplit
    | None
):
    """Run inference on input data, output predictions.

    Parameters
    ----------
    model
        Neural network model (e.g. pasa).
    datamodule
        The lightning DataModule to use for training **and** validation.
    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a lightning accelerator setup.

    Returns
    -------
    (
    list[BinaryPrediction]
    | list[MultiClassPrediction]
    | list[list[BinaryPrediction]]
    | list[list[MultiClassPrediction]]
    | BinaryPredictionSplit
    | MultiClassPredictionSplit
    | None
    )
        Depending on the return type of the DataModule's
        ``predict_dataloader()`` method:

        * if :py:class:`torch.utils.data.DataLoader`, then returns a
          :py:class:`list` of predictions.
        * if :py:class:`list` of :py:class:`torch.utils.data.DataLoader`, then
          returns a list of lists of predictions, each list corresponding to
          the iteration over one of the dataloaders.
        * if :py:class:`dict` of :py:class:`str` to
          :py:class:`torch.utils.data.DataLoader`, then returns a dictionary
          mapping names to lists of predictions.
        * if ``None``, then returns ``None``.

    Raises
    ------
    TypeError
        If the DataModule's ``predict_dataloader()`` method does not return any
        of the types described above.
    """

    from lightning.pytorch.loggers.logger import DummyLogger

    accelerator, devices = device_manager.lightning_accelerator()
    trainer = lightning.pytorch.Trainer(
        accelerator=accelerator,
        devices=devices,
        logger=DummyLogger(),
    )

    def _flatten(p: list[list]):
        return [sample for batch in p for sample in batch]

    dataloaders = datamodule.predict_dataloader()

    if isinstance(dataloaders, torch.utils.data.DataLoader):
        logger.info("Running prediction on a single dataloader...")
        return _flatten(trainer.predict(model, dataloaders))  # type: ignore

    if isinstance(dataloaders, list):
        retval_list = []
        for k, dataloader in enumerate(dataloaders):
            logger.info(f"Running prediction on split `{k}`...")
            retval_list.append(_flatten(trainer.predict(model, dataloader)))  # type: ignore
        return retval_list

    if isinstance(dataloaders, dict):
        retval_dict = {}
        for name, dataloader in dataloaders.items():
            logger.info(f"Running prediction on `{name}` split...")
            retval_dict[name] = _flatten(trainer.predict(model, dataloader))  # type: ignore
        return retval_dict

    if dataloaders is None:
        logger.warning("Datamodule did not return any prediction dataloaders!")
        return None

    # if you get to this point, then the user is returning something that is
    # not supported - complain!
    raise TypeError(
        f"Datamodule returned strangely typed prediction "
        f"dataloaders: `{type(dataloaders)}` - Please write code "
        f"to support this use-case.",
    )
