# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing

import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data
import torchvision.models as models
from mednet.libs.common.data.typing import TransformSequence
from mednet.libs.common.models.model import Model

from .separate import separate

logger = logging.getLogger("mednet")


class Densenet(Model):
    """Densenet-121 module.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    pretrained
        If set to True, loads pretrained model weights during initialization,
        else trains a new model.
    dropout
        Dropout rate after each dense layer.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: torch.nn.Module = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] = {},
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {},
        model_transforms: TransformSequence = [],
        augmentation_transforms: TransformSequence = [],
        pretrained: bool = False,
        dropout: float = 0.1,
        num_classes: int = 1,
    ):
        super().__init__(
            loss_type,
            loss_arguments,
            optimizer_type,
            optimizer_arguments,
            model_transforms,
            augmentation_transforms,
            num_classes,
        )

        self.name = "densenet-121"
        self.num_classes = num_classes

        self.pretrained = pretrained

        # Load pretrained model
        if not pretrained:
            weights = None
        else:
            logger.info(f"Loading pretrained {self.name} model weights")
            weights = models.DenseNet121_Weights.DEFAULT

        self.model_ft = models.densenet121(weights=weights, drop_rate=dropout)

        # Adapt output features
        self.model_ft.classifier = torch.nn.Linear(
            self.model_ft.classifier.in_features,
            self.num_classes,
        )

    def forward(self, x):
        x = self.normalizer(x)  # type: ignore
        return self.model_ft(x)

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initialize the normalizer for the current model.

        This function is NOOP if ``pretrained = True`` (normalizer set to
        imagenet weights, during contruction).

        Parameters
        ----------
        dataloader
            A torch Dataloader from which to compute the mean and std.
            Will not be used if the model is pretrained.
        """

        if self.pretrained:
            from .normalizer import make_imagenet_normalizer

            logger.warning(
                f"ImageNet pre-trained {self.name} model - NOT "
                f"computing z-norm factors from train dataloader. "
                f"Using preset factors from torchvision.",
            )
            self.normalizer = make_imagenet_normalizer()
        else:
            from .normalizer import make_z_normalizer

            logger.info(
                f"Uninitialised {self.name} model - "
                f"computing z-norm factors from train dataloader.",
            )
            self.normalizer = make_z_normalizer(dataloader)

    def training_step(self, batch, _):
        images = batch[0]
        labels = batch[1]["target"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        outputs = self(self.augmentation_transforms(images))

        return self._train_loss(outputs, labels.float())

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        images = batch[0]
        labels = batch[1]["target"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        outputs = self(images)

        return self._validation_loss(outputs, labels.float())

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        outputs = self(batch[0])
        probabilities = torch.sigmoid(outputs)
        return separate((probabilities, batch[1]))
