# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import lightning.pytorch as pl
import torch

from .separate import separate


class MultiLayerPerceptron(pl.LightningModule):
    """MLP with a variable number of inputs and hidden neurons (single layer).

    Parameters
    ----------
    train_loss
        The loss to be used during the training.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    validation_loss
        The loss to be used for validation (may be different from the training
        loss).  If extra-validation sets are provided, the same loss will be
        used throughout.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    input_size
        The number of inputs this classifer shall process.
    hidden_size
        The number of neurons on the single hidden layer.
    """

    def __init__(
        self,
        train_loss: torch.nn.Module = torch.nn.BCEWithLogitsLoss(),
        validation_loss: torch.nn.Module | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {"lr": 1e-2},
        input_size: int = 14,
        hidden_size: int = 10,
    ):
        super().__init__()

        self._train_loss = train_loss.to(self.device)
        self._validation_loss = (
            validation_loss if validation_loss is not None else train_loss
        ).to(self.device)
        self._optimizer_type = optimizer_type
        self._optimizer_arguments = optimizer_arguments

        self.name = "mlp"

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, 1)

    def forward(self, x):
        return self.fc2(self.relu(self.fc1(x)))

    def training_step(self, batch, batch_idx):
        _input = batch[1]
        labels = batch[2]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        output = self(_input)

        # Manually move criterion to selected device, since not part of the model.
        self._train_loss = self._train_loss.to(self.device)
        training_loss = self._train_loss(output, labels.float())

        return {"loss": training_loss}

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        _input = batch[1]
        labels = batch[2]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        output = self(_input)

        # Manually move criterion to selected device, since not part of the model.
        self._validation_loss = self._validation_loss.to(self.device)
        validation_loss = self._validation_loss(output, labels.float())

        if dataloader_idx == 0:
            return {"validation_loss": validation_loss}

        return {f"extra_validation_loss_{dataloader_idx}": validation_loss}

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        outputs = self(batch[0])
        probabilities = torch.sigmoid(outputs)
        return separate((probabilities, batch[1]))

    def configure_optimizers(self):
        return self._optimizer_type(
            self.parameters(),
            **self._optimizer_arguments,
        )
