# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click
from clapper.click import ResourceOption, verbosity_option
from clapper.logging import setup
from mednet.libs.common.scripts.click import ConfigCommand

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="mednet.libs.classification.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Run evaluation on an existing prediction output:

   .. code:: sh

      mednet classification evaluate -vv --predictions=path/to/predictions.json

2. Run evaluation on an existing prediction output, tune threshold a priori on the `validation` set:

   .. code:: sh

      mednet classification evaluate -vv --predictions=path/to/predictions.json --threshold=validation
""",
)
@click.option(
    "--predictions",
    "-p",
    help="Filename in which predictions are currently stored",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=True,
        path_type=pathlib.Path,
    ),
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""This value is used to define positives and negatives from
    probability outputs in predictions, and report performance measures on
    **binary** classification tasks. It should either come from the training
    set or a separate validation set to avoid biasing the analysis.
    Optionally, if you provide a multi-split set of predictions as input, this
    may also be the name of an existing split (e.g. ``validation``) from which
    the threshold will be estimated (by calculating the threshold leading to
    the highest F1-score on that set) and then applied to the subsequent
    sets.  This value is not used for multi-class classification tasks.""",
    default=0.5,
    show_default=True,
    required=True,
    type=click.STRING,
    cls=ResourceOption,
)
@click.option(
    "--binning",
    "-b",
    help="""The binning algorithm to use for computing the bin widths and
    distribution for histograms.  Choose from algorithms supported by
    :py:func:`numpy.histogram`, or a simple integer indicating the number of
    bins to have in the interval ``[0, 1]``.""",
    default="50",
    show_default=True,
    required=True,
    type=click.STRING,
    cls=ResourceOption,
)
@click.option(
    "--plot/--no-plot",
    "-P",
    help="""If set, then also produces figures containing the plots of
    performance curves and score histograms.""",
    required=True,
    show_default=True,
    default=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def evaluate(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    binning: str,
    plot: bool,
    **_,  # ignored
) -> None:  # numpydoc ignore=PR01
    """Evaluate predictions (from a model) on a classification task."""
    import json
    import typing

    from matplotlib.backends.backend_pdf import PdfPages
    from mednet.libs.common.scripts.utils import (
        execution_metadata,
        save_json_with_backup,
    )

    from ..engine.evaluator import (
        NumpyJSONEncoder,
        aggregate_pr,
        aggregate_roc,
        run_binary,
        score_plot,
        tabulate_results,
    )

    evaluation_filename = "evaluation.json"
    evaluation_file = pathlib.Path(output_folder) / evaluation_filename

    with predictions.open("r") as f:
        predict_data = json.load(f)

    # register metadata
    json_data: dict[str, typing.Any] = execution_metadata()
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(evaluation_file.with_suffix(".meta.json"), json_data)

    if threshold in predict_data:
        # it is the name of a split
        # first run evaluation for reference dataset
        from ..engine.evaluator import maxf1_threshold

        use_threshold = maxf1_threshold(predict_data[threshold])
        logger.info(f"Setting --threshold={use_threshold:.5f}")

    else:
        # we try to convert it to float and complain if that is not possible
        try:
            use_threshold = float(threshold)
        except ValueError:
            raise click.BadParameter(
                f"""The value of --threshold=`{threshold}` does not match one
                of the database split names ({', '.join(predict_data.keys())})
                or can not be converted to a float. Check your input.""",
            )

    results: dict[str, dict[str, typing.Any]] = dict()
    for k, v in predict_data.items():
        logger.info(f"Analyzing split `{k}`...")
        results[k] = run_binary(
            name=k,
            predictions=v,
            binning=int(binning) if binning.isnumeric() else binning,
            threshold_a_priori=use_threshold,
        )

    # records full result analysis to a JSON file
    logger.info(f"Saving evaluation results at `{evaluation_file}`...")
    with evaluation_file.open("w") as f:
        json.dump(results, f, indent=2, cls=NumpyJSONEncoder)

    # dump evaluation results in RST format to screen and file
    table_data = {}
    for k, v in results.items():
        table_data[k] = {
            kk: vv for kk, vv in v.items() if kk not in ("curves", "score-histograms")
        }
    table = tabulate_results(table_data, fmt="rst")
    click.echo(table)

    table_path = evaluation_file.with_suffix(".rst")
    logger.info(
        f"Saving evaluation results in table format at `{table_path}`...",
    )
    with table_path.open("w") as f:
        f.write(table)

    # dump evaluation plots in file
    figure_path = evaluation_file.with_suffix(".pdf")
    logger.info(f"Saving evaluation figures at `{figure_path}`...")

    if plot:
        with PdfPages(figure_path) as pdf:
            pr_curves = {k: v["curves"]["precision_recall"] for k, v in results.items()}
            pr_fig = aggregate_pr(pr_curves)
            pdf.savefig(pr_fig)

            roc_curves = {k: v["curves"]["roc"] for k, v in results.items()}
            roc_fig = aggregate_roc(roc_curves)
            pdf.savefig(roc_fig)

            # score plots
            for k, v in results.items():
                score_fig = score_plot(
                    v["score-histograms"],
                    f"Score distribution (split: {k})",
                    v["threshold"],
                )
                pdf.savefig(score_fig)
