# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
from datetime import datetime

import click
from clapper.click import ConfigCommand, ResourceOption, verbosity_option
from clapper.logging import setup

from .train import reusable_options as training_options

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="mednet.libs.classification.config",
    cls=ConfigCommand,
    epilog="""Examples:

  1. Train a pasa model with montgomery dataset, on the CPU, for only two
     epochs, then runs inference and evaluation on stock datasets, report
     performance as a table and figures:

     .. code:: sh

        $ mednet classification experiment -vv pasa montgomery --epochs=2
""",
)
@training_options
@verbosity_option(logger=logger, cls=ResourceOption)
@click.pass_context
def experiment(
    ctx,
    model,
    output_folder,
    epochs,
    batch_size,
    accumulate_grad_batches,
    drop_incomplete_batch,
    datamodule,
    validation_period,
    device,
    cache_samples,
    seed,
    parallel,
    monitoring_interval,
    augmentations,
    **_,
):  # numpydoc ignore=PR01
    r"""Run a complete experiment, from training, to prediction and evaluation.

    This script is just a wrapper around the individual scripts for training,
    running prediction, and evaluating.  It organises the output in a preset way:

    .. code::

       └─ <output-folder>/
          ├── model/  # the generated model will be here
          ├── predictions.json  # the prediction outputs for the sets
          └── evaluation/  # the outputs of the evaluations for the sets
    """

    experiment_start_timestamp = datetime.now()

    train_start_timestamp = datetime.now()
    logger.info(f"Started training at {train_start_timestamp}")

    from .train import train

    train_output_folder = output_folder / "model"
    ctx.invoke(
        train,
        model=model,
        output_folder=train_output_folder,
        epochs=epochs,
        batch_size=batch_size,
        accumulate_grad_batches=accumulate_grad_batches,
        drop_incomplete_batch=drop_incomplete_batch,
        datamodule=datamodule,
        validation_period=validation_period,
        device=device,
        cache_samples=cache_samples,
        seed=seed,
        parallel=parallel,
        monitoring_interval=monitoring_interval,
        augmentations=augmentations,
    )
    train_stop_timestamp = datetime.now()

    logger.info(f"Ended training at {train_stop_timestamp}")
    logger.info(f"Training runtime: {train_stop_timestamp-train_start_timestamp}")

    logger.info("Started train analysis")
    from mednet.libs.common.scripts.train_analysis import train_analysis

    logdir = train_output_folder / "logs"
    ctx.invoke(
        train_analysis,
        logdir=logdir,
        output_folder=train_output_folder,
    )

    logger.info("Ended train analysis")

    predict_start_timestamp = datetime.now()
    logger.info(f"Started prediction at {predict_start_timestamp}")

    from .predict import predict

    predictions_output = output_folder / "predictions"

    ctx.invoke(
        predict,
        output_folder=predictions_output,
        model=model,
        datamodule=datamodule,
        device=device,
        weight=train_output_folder,
        batch_size=batch_size,
        parallel=parallel,
    )

    predict_stop_timestamp = datetime.now()
    logger.info(f"Ended prediction at {predict_stop_timestamp}")
    logger.info(f"Prediction runtime: {predict_stop_timestamp-predict_start_timestamp}")

    evaluation_start_timestamp = datetime.now()
    logger.info(f"Started evaluation at {evaluation_start_timestamp}")

    from .evaluate import evaluate

    predictions_file = predictions_output / "predictions.json"

    with (predictions_output / "predictions.json").open() as pf:
        splits = json.load(pf).keys()

        if "validation" in splits:
            evaluation_threshold = "validation"
        elif "train" in splits:
            evaluation_threshold = "train"
        else:
            evaluation_threshold = None

    ctx.invoke(
        evaluate,
        predictions=predictions_file,
        output_folder=output_folder,
        threshold=evaluation_threshold,
    )

    evaluation_stop_timestamp = datetime.now()
    logger.info(f"Ended prediction at {evaluation_stop_timestamp}")
    logger.info(
        f"Prediction runtime: {evaluation_stop_timestamp-evaluation_start_timestamp}"
    )

    saliency_map_generation_start_timestamp = datetime.now()
    logger.info(
        f"Started saliency map generation at {saliency_map_generation_start_timestamp}"
    )

    from .saliency.generate import generate

    saliencies_gen_folder = output_folder / "gradcam" / "saliencies"

    ctx.invoke(
        generate,
        model=model,
        datamodule=datamodule,
        weight=train_output_folder,
        output_folder=saliencies_gen_folder,
    )

    saliency_map_generation_stop_timestamp = datetime.now()
    logger.info(
        f"Ended saliency map generation at {saliency_map_generation_stop_timestamp}"
    )
    logger.info(
        f"Saliency map generation runtime: {saliency_map_generation_stop_timestamp-saliency_map_generation_start_timestamp}"
    )

    saliency_images_generation_start_timestamp = datetime.now()
    logger.info(
        f"Started generating saliency images at {saliency_images_generation_start_timestamp}"
    )

    from .saliency.view import view

    saliencies_view_folder = output_folder / "gradcam" / "visualizations"

    ctx.invoke(
        view,
        model=model,
        datamodule=datamodule,
        input_folder=saliencies_gen_folder,
        output_folder=saliencies_view_folder,
    )

    saliency_images_generation_stop_timestamp = datetime.now()
    logger.info(
        f"Ended saliency images generation at {saliency_images_generation_stop_timestamp}"
    )
    logger.info(
        f"Saliency images generation runtime: {saliency_images_generation_stop_timestamp-saliency_images_generation_start_timestamp}"
    )

    experiment_stop_timestamp = datetime.now()
    logger.info(
        f"Total experiment runtime: {experiment_stop_timestamp-experiment_start_timestamp}"
    )
