# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later


import click
from clapper.click import ResourceOption, verbosity_option
from clapper.logging import setup
from mednet.libs.common.scripts.click import ConfigCommand
from mednet.libs.common.scripts.predict import reusable_options

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="mednet.libs.classification.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Run prediction on an existing DataModule configuration:

   .. code:: sh

      mednet classification predict -vv pasa montgomery --weight=path/to/model.ckpt --output=path/to/predictions.json

2. Enable multi-processing data loading with 6 processes:

   .. code:: sh

      mednet classification predict -vv pasa montgomery --parallel=6 --weight=path/to/model.ckpt --output=path/to/predictions.json

""",
)
@reusable_options
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def predict(
    output_folder,
    model,
    datamodule,
    batch_size,
    device,
    weight,
    parallel,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Run inference (generates scores) on all input images, using a pre-trained model."""

    import json

    from mednet.libs.classification.engine.predictor import run
    from mednet.libs.common.engine.device import DeviceManager
    from mednet.libs.common.scripts.predict import (
        load_checkpoint,
        save_json_data,
        setup_datamodule,
    )
    from mednet.libs.common.scripts.utils import save_json_with_backup

    predictions_meta_file = output_folder / "predictions.meta.json"
    predictions_meta_file.parent.mkdir(parents=True, exist_ok=True)

    setup_datamodule(datamodule, model, batch_size, parallel)
    model = load_checkpoint(model, weight)
    device_manager = DeviceManager(device)
    save_json_data(datamodule, model, device_manager, predictions_meta_file)

    predictions = run(model, datamodule, device_manager)

    predictions_file = output_folder / "predictions.json"
    save_json_with_backup(predictions_file, predictions)

    with predictions_file.open("w") as f:
        json.dump(predictions, f, indent=2)
    logger.info(f"Predictions saved to `{str(predictions_file)}`")
