# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import configparser
import logging
import pathlib
import shutil

import gitlab

logger = logging.getLogger(__name__)


def gitlab_instance_and_token() -> tuple[gitlab.Gitlab, str]:
    """Return an instance of the Gitlab object for remote operations, and the
    user token.

    Returns
    -------
        Gitlab main object and user token
    """

    cfg = pathlib.Path("~/.python-gitlab.cfg").expanduser()
    if cfg.exists():
        gl = gitlab.Gitlab.from_config("idiap", [str(cfg)])
        config = configparser.ConfigParser()
        config.read(cfg)
        token = config["idiap"]["private_token"]

    else:  # ask the user for a token or use one from the current runner
        server = "https://gitlab.idiap.ch"
        token = input(f"{server} (user or project) token: ")
        gl = gitlab.Gitlab(server, private_token=token, api_version="4")

    # tests authentication with given credential.
    gl.auth()

    return gl, token


def sanitize_filename(tmpdir: pathlib.Path, path: pathlib.Path) -> pathlib.Path:
    """Sanitize the name of a file to be logged.

    This function sanitizes the basename of a file to be logged on the GitLab
    MLflow server. It removes unsupported characters (such as ``=``) by
    creating a copy of the file to be uploaded, with a modified name, on the
    provided temporary directory.  It then returns the name of such a temporary
    file.

    If the input file path does not need sanitization, it is returned as is.

    Parameters
    ----------
    tmpdir
        The temporary directory where a copy of the input path, with a
        sanitized name will be created.
    path
        The file that needs its name sanitized.

    Returns
    -------
        Path to the temporary folder, and the sanitized copy of the input file
        in said temporary folder, or the input ``path``, in case its name does not
        need sanitization.
    """

    sanitized_filename = path.parts[-1].replace("=", "-")
    if path.parts[-1] == sanitized_filename:
        return path

    absolute_sanitized_filename = tmpdir / sanitized_filename
    logger.info(f"Sanitazing filename `{path}` -> `{absolute_sanitized_filename}`")
    shutil.copy2(path, absolute_sanitized_filename)
    return absolute_sanitized_filename


def size_in_mb(path: pathlib.Path) -> float:
    """Return the size in megabytes of a file.

    Parameters
    ----------
    path
        Input path to calculate file size from.

    Returns
    -------
        A floating point number for the size of the object in MB.
    """
    return path.stat().st_size / (1024**2)
