# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing

import lightning.pytorch as pl
import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data
import torchvision.transforms
from mednet.libs.common.data.typing import TransformSequence

from .loss_weights import get_positive_weights
from .typing import Checkpoint

logger = logging.getLogger("mednet")


class Model(pl.LightningModule):
    """Base class for models.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: torch.nn.Module = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] = {},
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {},
        model_transforms: TransformSequence = [],
        augmentation_transforms: TransformSequence = [],
        num_classes: int = 1,
    ):
        super().__init__()

        self.name = "model"
        self.num_classes = num_classes

        self.model_transforms = model_transforms

        self._loss_type = loss_type

        self._train_loss = None
        self._train_loss_arguments = loss_arguments

        self._validation_loss = None
        self._validation_loss_arguments = loss_arguments

        self._optimizer_type = optimizer_type
        self._optimizer_arguments = optimizer_arguments

        self.augmentation_transforms = augmentation_transforms

    @property
    def augmentation_transforms(self) -> torchvision.transforms.Compose:
        return self._augmentation_transforms

    @augmentation_transforms.setter
    def augmentation_transforms(self, v: TransformSequence):
        self._augmentation_transforms = torchvision.transforms.Compose(v)

        if len(v) != 0:
            transforms_str = ", ".join(
                [
                    f"{type(k).__module__}.{str(k)}"
                    for k in self._augmentation_transforms.transforms
                ]
            )
            logger.info(f"Data augmentations: {transforms_str}")
        else:
            logger.info("Data augmentations: None")

    def forward(self, x):
        raise NotImplementedError

    def on_save_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Perform actions during checkpoint saving (called by lightning).

        Called by Lightning when saving a checkpoint to give you a chance to
        store anything else you might want to save. Use on_load_checkpoint() to
        restore what additional data is saved here.

        Parameters
        ----------
        checkpoint
            The checkpoint to save.
        """

        checkpoint["normalizer"] = self.normalizer

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Perform actions during model loading (called by lightning).

        If you saved something with on_save_checkpoint() this is your chance to
        restore this.

        Parameters
        ----------
        checkpoint
            The loaded checkpoint.
        """

        logger.info("Restoring normalizer from checkpoint.")
        self.normalizer = checkpoint["normalizer"]

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initialize the input normalizer for the current model.

        Parameters
        ----------
        dataloader
            A torch Dataloader from which to compute the mean and std.
        """

        from .normalizer import make_z_normalizer

        logger.info(
            f"Uninitialised {self.name} model - "
            f"computing z-norm factors from train dataloader.",
        )
        self.normalizer = make_z_normalizer(dataloader)

    def training_step(self, batch, _):
        raise NotImplementedError

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        raise NotImplementedError

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        raise NotImplementedError

    def configure_losses(self):
        self._train_loss = self._loss_type(**self._train_loss_arguments)
        self._validation_loss = self._loss_type(**self._validation_loss_arguments)

    def configure_optimizers(self):
        return self._optimizer_type(
            self.parameters(),
            **self._optimizer_arguments,
        )

    def to(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Self:
        """Move model, augmentations and losses to specified device.

        Refer to the method :py:meth:`torch.nn.Module.to` for details.

        Parameters
        ----------
        *args
            Parameter forwarded to the underlying implementations.
        **kwargs
            Parameter forwarded to the underlying implementations.

        Returns
        -------
            Self.
        """

        super().to(*args, **kwargs)

        self._augmentation_transforms = torchvision.transforms.Compose(
            [
                k.to(*args, **kwargs)
                for k in self._augmentation_transforms.transforms
                if hasattr(k, "to")
            ]
        )

        if self._train_loss is not None:
            self._train_loss.to(*args, **kwargs)
        if self._validation_loss is not None:
            self._validation_loss.to(*args, **kwargs)

        return self

    def balance_losses(self, datamodule) -> None:
        """Balance the loss based on the distribution of targets in the datamodule, if the loss supports it (contains a 'pos_weight' attribute).

        Parameters
        ----------
        datamodule
            Instance of a datamodule.
        """

        try:
            getattr(self._loss_type(), "pos_weight")
        except AttributeError:
            logger.warning(
                f"Loss {self._loss_type} does not posess a 'pos_weight' attribute and will not be balanced."
            )
        else:
            logger.info(f"Balancing training loss {self._loss_type}.")
            train_weights = get_positive_weights(datamodule.train_dataloader())
            self._train_loss_arguments["pos_weight"] = train_weights

            logger.info(f"Balancing validation loss {self._loss_type}.")
            if "validation" in datamodule.val_dataloader().keys():
                validation_weights = get_positive_weights(
                    datamodule.val_dataloader()["validation"]
                )
            else:
                logger.warning(
                    "Datamodule does not contain a validation dataloader. The training dataloader will be used instead."
                )
                validation_weights = get_positive_weights(datamodule.train_dataloader())
            self._validation_loss_arguments["pos_weight"] = validation_weights
