# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import torch
import torch.nn
import torchvision.transforms.functional


def crop_image_to_mask(img: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """Square crop image to the boundaries of a boolean mask.

    Parameters
    ----------
    img
        The image to crop, of shape channels x height x width.
    mask
        The boolean mask to use for cropping.

    Returns
    -------
        The cropped image.
    """

    if img.shape[-2:] != mask.shape[-2:]:
        raise ValueError(
            f"Image and mask must have the same size: {img.shape[-2:]} != {mask.shape[-2:]}"
        )

    h, w = img.shape[-2:]

    flat_mask = mask.flatten()
    top = flat_mask.nonzero()[0] // w
    bottom = h - (torch.flip(flat_mask, dims=(0,)).nonzero()[0] // w)

    flat_transposed_mask = torch.transpose(mask, 1, 2).flatten()
    left = flat_transposed_mask.nonzero()[0] // h
    right = w - (torch.flip(flat_transposed_mask, dims=(0,)).nonzero()[0] // h)

    return img[:, top:bottom, left:right]


def crop_multiple_images_to_mask(
    images: list[torch.Tensor], mask: torch.Tensor
) -> list[torch.Tensor]:
    """Apply crop_images_to_mask on multiple images.

    Parameters
    ----------
    images
        List of images to crop, of shape channels x height x width.
    mask
        The boolean mask to use for cropping.

    Returns
    -------
        A list of cropped images.
    """
    return [crop_image_to_mask(img, mask) for img in images]


def resize_max_side(tensor: torch.Tensor, max_side: int) -> torch.Tensor:
    """Resize image based on the longest side while keeping the aspect ratio.

    Parameters
    ----------
    tensor
        The tensor to resize.
    max_side
        The new length of the largest side.

    Returns
    -------
        The resized image.
    """

    from torchvision.transforms import InterpolationMode

    if max_side <= 0:
        raise ValueError(f"The new max side ({max_side}) must be positive.")

    height, width = tensor.shape[-2:]
    aspect_ratio = float(height) / float(width)

    if height >= width:
        new_size = (max_side, int(max_side / aspect_ratio))
    else:
        new_size = (int(max_side * aspect_ratio), max_side)

    return torchvision.transforms.Resize(
        new_size, interpolation=InterpolationMode.NEAREST, antialias=True
    )(tensor)


def square_center_pad(img: torch.Tensor) -> torch.Tensor:
    """Return a squared version of the image, centered on a canvas padded with
    zeros.

    Parameters
    ----------
    img
        The tensor to be transformed.  Expected to be in the form: ``[...,
        [1,3], H, W]`` (i.e. arbitrary number of leading dimensions).

    Returns
    -------
        Transformed tensor, guaranteed to be square (ie. equal height and
        width).
    """

    height, width = img.shape[-2:]
    maxdim = numpy.max([height, width])

    # padding
    left = (maxdim - width) // 2
    top = (maxdim - height) // 2
    right = maxdim - width - left
    bottom = maxdim - height - top

    return torchvision.transforms.functional.pad(
        img,
        [left, top, right, bottom],
        0,
        "constant",
    )


def grayscale_to_rgb(img: torch.Tensor) -> torch.Tensor:
    """Convert an image in grayscale to RGB.

    If the image is already in RGB format, then this is a NOOP - the same
    tensor is returned (no cloning).  If the image is in grayscale format
    (number of color channels = 1), then replicate it to obtain 3 color channels
    (a new copy is returned in this case).

    Parameters
    ----------
    img
        The tensor to be transformed.  Expected to be in the form: ``[...,
        [1,3], H, W]`` (i.e. arbitrary number of leading dimensions).

    Returns
    -------
    torch.Tensor
        Transformed tensor with 3 identical color channels.
    """

    if img.ndim < 3:
        raise TypeError(
            f"Input image tensor should have at least 3 dimensions, "
            f"but found {img.ndim}. If a grayscale image was provided, "
            f"ensure to include a channel dimension of size 1 ( i.e: "
            f"[1, height, width]).",
        )

    if img.shape[-3] not in (1, 3):
        raise TypeError(
            f"Input image tensor should have 1 or 3 color channels,"
            f"but found {img.shape[-3]}.",
        )

    if img.shape[-3] == 3:
        return img

    # it is a grayscale image - repeat the image 3 times
    repetitions = img.dim() * [1]
    repetitions[-3] = 3
    return img.repeat(*repetitions)


def rgb_to_grayscale(img: torch.Tensor) -> torch.Tensor:
    """Convert an image in RGB to grayscale.

    If the image is already in grayscale format, then this is a NOOP - the same
    tensor is returned (no cloning).  If the image is in RGB format
    (number of color channels = 3), then compresses the color channels into
    a single grayscale channel following this equation:

    .. math::

       grayscale = (0.2989 * r + 0.587 * g + 0.114 * b)

    A new tensor is returned in this case.

    Parameters
    ----------
    img
        The tensor to be transformed.  Expected to be in the form: ``[...,
        [1,3], H, W]`` (i.e. arbitrary number of leading dimensions).

    Returns
    -------
    torch.Tensor
        Transformed tensor with a single (grayscale) color channel.
    """

    if img.ndim < 3:
        raise TypeError(
            f"Input image tensor should have at least 3 dimensions, "
            f"but found {img.ndim}. If a grayscale image was provided, "
            f"ensure to include a channel dimension of size 1 ( i.e: "
            f"[1, height, width]).",
        )

    if img.shape[-3] not in (1, 3):
        raise TypeError(
            f"Input image tensor should have 1 or 3 planes,"
            f"but found {img.shape[-3]}",
        )

    if img.shape[-3] == 1:
        return img

    # it is an RGB image - use torchvision implementation
    return torchvision.transforms.functional.rgb_to_grayscale(img)


class ResizeMaxSide(torch.nn.Module):
    """Resize image on the longest side while keeping the aspect ratio.

    Parameters
    ----------
    max_side
        The new length of the largest side.
    """

    def __init__(self, max_side: int):
        super().__init__()
        self.max_side = max_side

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return resize_max_side(img, self.max_side)


class SquareCenterPad(torch.nn.Module):
    """Transform to a squared version of the image, centered on a canvas padded
    with zeros.
    """

    def __init__(self):
        super().__init__()

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return square_center_pad(img)


class RGB(torch.nn.Module):
    """Wrapper class around :py:func:`.grayscale_to_rgb` to be used as a model
    transform.
    """

    def __init__(self):
        super().__init__()

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return grayscale_to_rgb(img)


class Grayscale(torch.nn.Module):
    """Wrapper class around :py:func:`rgb_to_grayscale` to be used as a model
    transform.
    """

    def __init__(self):
        super().__init__()

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return rgb_to_grayscale(img)
