# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def list_(config: dict[str, dict[str, str]]) -> None:
    """List all supported and configured databases.

    Parameters
    ----------
    config
        Dictionary where keys are database names, and values are dictionaries
        containing two string keys:

        * ``module``: the full Pythonic module name (e.g.
        ``mednet.libs.classification.data.montgomery``).
        * ``datadir``: points to the user-configured data directory for the
        current dataset, if set, or ``None`` otherwise.
    """

    click.echo("Available databases:")
    for k, v in config.items():
        if "datadir" not in v:
            # this database does not have a "datadir"
            continue

        if v["datadir"] is not None:
            click.secho(f'- {k} ({v["module"]}): "{v["datadir"]}"', fg="green")
        else:
            click.echo(f'- {k} ({v["module"]}): NOT installed')


def check(entry_point_group, fold, limit):  # numpydoc ignore=PR01
    """Check file access on one or more DataModules."""
    import importlib.metadata
    import sys

    click.secho(f"Checking fold `{fold}`...", fg="yellow")
    try:
        module = importlib.metadata.entry_points(group=entry_point_group)[fold].module
    except KeyError:
        raise Exception(f"Could not find database fold `{fold}`")

    datamodule = importlib.import_module(module).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.batch_size = 1  # ensure one sample is loaded at a time
    datamodule.setup("predict")  # sets up all datasets

    loaders = datamodule.predict_dataloader()

    errors = 0
    for k, loader in loaders.items():
        if limit == 0:
            click.secho(
                f"Checking all samples of dataset `{k}` at fold `{fold}`...",
                fg="yellow",
            )
            loader_limit = sys.maxsize
        else:
            click.secho(
                f"Checking first {limit} samples of dataset "
                f"`{k}` at fold `{fold}`...",
                fg="yellow",
            )
            loader_limit = limit
        # the for loop will trigger raw data loading (ie. user code), protect
        # it
        try:
            for i, batch in enumerate(loader):
                if loader_limit == 0:
                    break
                logger.info(
                    f"{batch[1]['name'][0]}: "
                    f"{[s for s in batch[0][0].shape]}@{batch[0][0].dtype}",
                )
                loader_limit -= 1
        except Exception:
            logger.exception(f"Unable to load batch {i} in dataset {k}")
            errors += 1

    if not errors:
        click.secho(
            f"OK! No errors were reported for database fold `{fold}`.",
            fg="green",
        )
    else:
        click.secho(
            f"Found {errors} errors loading DataModule `{fold}`.",
            fg="red",
        )
