# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import pathlib
import typing

import click
import mednet.libs.common.data.datamodule
import mednet.libs.common.models.model
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def reusable_options(f: typing.Callable):
    """Wrap reusable predict script options for other scripts.

    This decorator equips the target function ``f`` with all (reusable)
    ``predict`` script options.

    Parameters
    ----------
    f
        The target function to equip with options.  This function must have
        parameters that accept such options.

    Returns
    -------
        The decorated version of function ``f``
    """
    from clapper.click import ResourceOption

    @click.option(
        "--output-folder",
        "-o",
        help="Directory in which to save predictions (created if does not exist)",
        required=True,
        type=click.Path(
            file_okay=False,
            dir_okay=True,
            writable=True,
            path_type=pathlib.Path,
        ),
        default="predictions",
        cls=ResourceOption,
    )
    @click.option(
        "--model",
        "-m",
        help="""A lightning module instance implementing the network architecture
        (not the weights, necessarily) to be used for prediction.""",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--datamodule",
        "-d",
        help="""A lightning DataModule that will be asked for prediction data
        loaders. Typically, this includes all configured splits in a DataModule,
        however this is not a requirement.  A DataModule that returns a single
        dataloader for prediction (wrapped in a dictionary) is acceptable.""",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--batch-size",
        "-b",
        help="""Number of samples in every batch (this parameter affects memory
        requirements for the network).""",
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--device",
        "-d",
        help='A string indicating the device to use (e.g. "cpu" or "cuda:0")',
        show_default=True,
        required=True,
        default="cpu",
        cls=ResourceOption,
    )
    @click.option(
        "--weight",
        "-w",
        help="""Path or URL to pretrained model file (`.ckpt` extension),
        corresponding to the architecture set with `--model`.  Optionally, you may
        also pass a directory containing the result of a training session, in which
        case either the best (lowest validation) or latest model will be loaded.""",
        required=True,
        cls=ResourceOption,
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=True,
            readable=True,
            path_type=pathlib.Path,
        ),
    )
    @click.option(
        "--parallel",
        "-P",
        help="""Use multiprocessing for data loading: if set to -1 (default),
        disables multiprocessing data loading.  Set to 0 to enable as many data
        loading instances as processing cores available in the system.  Set to
        >= 1 to enable that many multiprocessing instances for data loading.""",
        type=click.IntRange(min=-1),
        show_default=True,
        required=True,
        default=-1,
        cls=ResourceOption,
    )
    @functools.wraps(f)
    def wrapper_reusable_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_reusable_options


def setup_datamodule(
    datamodule: mednet.libs.common.data.datamodule.ConcatDataModule,
    model: mednet.libs.common.models.model.Model,
    batch_size: int,
    parallel: int,
) -> None:  # numpydoc ignore=PR01
    """Configure and set up the datamodule."""

    datamodule.batch_size = batch_size
    datamodule.parallel = parallel
    datamodule.model_transforms = list(model.model_transforms)

    datamodule.prepare_data()
    datamodule.setup(stage="predict")


def load_checkpoint(
    model: mednet.libs.common.models.model.Model, weight: pathlib.Path
) -> mednet.libs.common.models.model.Model:
    """Load a model checkpoint for prediction.

    Parameters
    ----------
    model
        Instance of a model.
    weight
        The base directory containing either the "best", "last" or "periodic"
        checkpoint to start the training session from.

    Returns
    -------
        An instance of the model loaded from the checkpoint.
    """
    from mednet.libs.common.utils.checkpointer import (
        get_checkpoint_to_run_inference,
    )

    if weight.is_dir():
        weight = get_checkpoint_to_run_inference(weight)

    logger.info(f"Loading checkpoint from `{weight}`...")
    return type(model).load_from_checkpoint(weight, strict=False)


def save_json_data(
    datamodule: mednet.libs.common.data.datamodule.ConcatDataModule,
    model: mednet.libs.common.models.model.Model,
    device_manager,
    output_file: pathlib.Path,
) -> None:  # numpydoc ignore=PR01
    """Save prediction hyperparameters into a .json file."""

    from .utils import (
        device_properties,
        execution_metadata,
        model_summary,
        save_json_with_backup,
    )

    json_data: dict[str, typing.Any] = execution_metadata()
    json_data.update(device_properties(device_manager.device_type))
    json_data.update(
        dict(
            database_name=datamodule.database_name,
            database_split=datamodule.split_name,
            model_name=model.name,
        ),
    )
    json_data.update(model_summary(model))
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(output_file, json_data)
