# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import pathlib
import typing

import click
from clapper.click import ResourceOption
from clapper.logging import setup

# logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")
logger = setup("mednet", format="%(levelname)s: %(message)s")


def reusable_options(f):
    """Wrap reusable training script options (for ``experiment``).

    This decorator equips the target function ``f`` with all (reusable)
    ``train`` script options.

    Parameters
    ----------
    f
        The target function to equip with options.  This function must have
        parameters that accept such options.

    Returns
    -------
        The decorated version of function ``f``
    """

    @click.option(
        "--output-folder",
        "-o",
        help="Directory in which to store results (created if does not exist)",
        required=True,
        type=click.Path(
            file_okay=False,
            dir_okay=True,
            writable=True,
            path_type=pathlib.Path,
        ),
        default="results",
        cls=ResourceOption,
    )
    @click.option(
        "--model",
        "-m",
        help="A lightning module instance implementing the network to be trained",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--datamodule",
        "-d",
        help="A lightning DataModule containing the training and validation sets.",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--batch-size",
        "-b",
        help="Number of samples in every batch (this parameter affects "
        "memory requirements for the network).  If the number of samples in "
        "the batch is larger than the total number of samples available for "
        "training, this value is truncated.  If this number is smaller, then "
        "batches of the specified size are created and fed to the network "
        "until there are no more new samples to feed (epoch is finished).  "
        "If the total number of training samples is not a multiple of the "
        "batch-size, the last batch will be smaller than the first, unless "
        "--drop-incomplete-batch is set, in which case this batch is not used.",
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--accumulate-grad-batches",
        "-a",
        help="Number of accumulations for backward propagation to accumulate "
        "gradients over k batches before stepping the optimizer. This "
        "parameter, used in conjunction with the batch-size, may be used to "
        "reduce the number of samples loaded in each iteration, to affect memory "
        "usage in exchange for processing time (more iterations). This is "
        "useful interesting when one is training on GPUs with a limited amount "
        "of onboard RAM. The default of 1 forces the whole batch to be "
        "processed at once. Otherwise the batch is multiplied by "
        "accumulate-grad-batches pieces, and gradients are accumulated "
        "to complete each training step.",
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--drop-incomplete-batch/--no-drop-incomplete-batch",
        "-D",
        help="If set, the last batch in an epoch will be dropped if "
        "incomplete.  If you set this option, you should also consider "
        "increasing the total number of epochs of training, as the total number "
        "of training steps may be reduced.",
        required=True,
        show_default=True,
        default=False,
        cls=ResourceOption,
    )
    @click.option(
        "--epochs",
        "-e",
        help="""Number of epochs (complete training set passes) to train for.
        If continuing from a saved checkpoint, ensure to provide a greater
        number of epochs than was saved in the checkpoint to be loaded.""",
        show_default=True,
        required=True,
        default=1000,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--validation-period",
        "-p",
        help="""Number of epochs after which validation happens.  By default,
        we run validation after every training epoch (period=1).  You can
        change this to make validation more sparse, by increasing the
        validation period. Notice that this affects checkpoint saving.  While
        checkpoints are created after every training step (the last training
        step always triggers the overriding of latest checkpoint), and
        this process is independent of validation runs, evaluation of the
        'best' model obtained so far based on those will be influenced by this
        setting.""",
        show_default=True,
        required=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--device",
        "-x",
        help='A string indicating the device to use (e.g. "cpu" or "cuda:0")',
        show_default=True,
        required=True,
        default="cpu",
        cls=ResourceOption,
    )
    @click.option(
        "--cache-samples/--no-cache-samples",
        help="If set to True, loads the sample into memory, "
        "otherwise loads them at runtime.",
        required=True,
        show_default=True,
        default=False,
        cls=ResourceOption,
    )
    @click.option(
        "--seed",
        "-s",
        help="Seed to use for the random number generator",
        show_default=True,
        required=False,
        default=42,
        type=click.IntRange(min=0),
        cls=ResourceOption,
    )
    @click.option(
        "--parallel",
        "-P",
        help="""Use multiprocessing for data loading: if set to -1 (default),
        disables multiprocessing data loading.  Set to 0 to enable as many data
        loading instances as processing cores available in the system.  Set to
        >= 1 to enable that many multiprocessing instances for data
        loading.""",
        type=click.IntRange(min=-1),
        show_default=True,
        required=True,
        default=-1,
        cls=ResourceOption,
    )
    @click.option(
        "--monitoring-interval",
        "-I",
        help="""Time between checks for the use of resources during each training
        epoch, in seconds.  An interval of 5 seconds, for example, will lead to
        CPU and GPU resources being probed every 5 seconds during each training
        epoch. Values registered in the training logs correspond to averages
        (or maxima) observed through possibly many probes in each epoch.
        Notice that setting a very small value may cause the probing process to
        become extremely busy, potentially biasing the overall perception of
        resource usage.""",
        type=click.FloatRange(min=0.1),
        show_default=True,
        required=True,
        default=5.0,
        cls=ResourceOption,
    )
    @click.option(
        "--augmentations",
        "-A",
        help="""Models that can be trained in this package are shipped without
        explicit data augmentations. This option allows you to define a list of
        data augmentations to use for training the selected model.""",
        type=click.UNPROCESSED,
        default=[],
        cls=ResourceOption,
    )
    @functools.wraps(f)
    def wrapper_reusable_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_reusable_options


def get_checkpoint_file(results_dir) -> pathlib.Path:
    """Return the path of the latest checkpoint if it exists.

    Parameters
    ----------
    results_dir
        Directory in which results are saved.

    Returns
    -------
        Path to the latest checkpoint
    """
    from mednet.libs.common.utils.checkpointer import (
        get_checkpoint_to_resume_training,
    )

    checkpoint_file = None
    if results_dir.is_dir():
        try:
            checkpoint_file = get_checkpoint_to_resume_training(results_dir)
        except FileNotFoundError:
            logger.info(
                f"Folder {results_dir} already exists, but I did not"
                f" find any usable checkpoint file to resume training"
                f" from. Starting from scratch...",
            )

    return checkpoint_file


def load_checkpoint(checkpoint_file, datamodule, model):
    """Load the checkpoint.

    Parameters
    ----------
    checkpoint_file
        Path to the checkpoint.
    datamodule
        Instance of a Datamodule, used to set the model's normalizer.
    model
        The model corresponding to the checkpoint.
    """

    import torch

    if checkpoint_file is None or not hasattr(model, "on_load_checkpoint"):
        # Sets the model normalizer with the unaugmented-train-subset if we are
        # starting from scratch and/or the model does not contain its own
        # checkpoint loading strategy (e.g. a pytorch stock checkpoint). This
        # call may be a NOOP, if the model comes from outside this framework,
        # and expects different weights for the normalisation layer.
        if hasattr(model, "set_normalizer"):
            model.set_normalizer(datamodule.unshuffled_train_dataloader())
        else:
            logger.warning(
                f"Model {model.name} has no `set_normalizer` method. "
                "Skipping normalization setup (unsupported external model).",
            )
    else:
        # Normalizer will be loaded during model.on_load_checkpoint
        checkpoint = torch.load(checkpoint_file)
        start_epoch = checkpoint["epoch"]
        logger.info(
            f"Resuming from epoch {start_epoch} "
            f"(checkpoint file: `{str(checkpoint_file)}`)...",
        )


def setup_datamodule(
    datamodule,
    model,
    batch_size,
    drop_incomplete_batch,
    cache_samples,
    parallel,
) -> None:  # numpydoc ignore=PR01
    """Configure and set up the datamodule."""
    datamodule.batch_size = batch_size
    datamodule.drop_incomplete_batch = drop_incomplete_batch
    datamodule.cache_samples = cache_samples
    datamodule.parallel = parallel
    datamodule.model_transforms = model.model_transforms

    datamodule.prepare_data()
    datamodule.setup(stage="fit")


def save_json_data(
    datamodule,
    model,
    output_folder,
    device_manager,
    epochs,
    batch_size,
    accumulate_grad_batches,
    drop_incomplete_batch,
    validation_period,
    cache_samples,
    seed,
    parallel,
    monitoring_interval,
    **kwargs,
) -> None:  # numpydoc ignore=PR01
    """Save training hyperparameters into a .json file."""
    from .utils import (
        device_properties,
        execution_metadata,
        model_summary,
        save_json_with_backup,
    )

    json_data: dict[str, typing.Any] = execution_metadata()
    json_data.update(device_properties(device_manager.device_type))
    json_data.update(
        dict(
            database_name=datamodule.database_name,
            split_name=datamodule.split_name,
            epochs=epochs,
            batch_size=batch_size,
            accumulate_grad_batches=accumulate_grad_batches,
            drop_incomplete_batch=drop_incomplete_batch,
            validation_period=validation_period,
            cache_samples=cache_samples,
            seed=seed,
            parallel=parallel,
            monitoring_interval=monitoring_interval,
            model_name=model.name,
        ),
    )
    json_data.update(kwargs)
    json_data.update(model_summary(model))
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(output_folder / "meta.json", json_data)
