# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import pathlib

import click
from clapper.click import ResourceOption
from clapper.logging import setup

# logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")
logger = setup("mednet", format="%(levelname)s: %(message)s")


def reusable_options(f):
    """Wrap reusable upload script options.

    This decorator equips the target function ``f`` with all (reusable)
    ``upload`` script options.

    Parameters
    ----------
    f
        The target function to equip with options.  This function must have
        parameters that accept such options.

    Returns
    -------
        The decorated version of function ``f``
    """

    @click.option(
        "--project-path",
        "-p",
        help="Path to the project where to upload model entries",
        required=True,
        type=str,
        default="biosignal/software/mednet",
        show_default=True,
        cls=ResourceOption,
    )
    @click.option(
        "--experiment-folder",
        "-f",
        help="Directory in which to upload results from",
        required=True,
        type=click.Path(
            file_okay=False,
            dir_okay=True,
            path_type=pathlib.Path,
        ),
        default="results",
        show_default=True,
        cls=ResourceOption,
    )
    @click.option(
        "--experiment-name",
        "-e",
        help='A string indicating the experiment name (e.g. "exp-pasa-mc" or "exp-densenet-mc-ch")',
        cls=ResourceOption,
    )
    @click.option(
        "--run-name",
        "-r",
        help='A string indicating the run name (e.g. "run-1")',
        cls=ResourceOption,
    )
    @click.option(
        "--upload-limit-mb",
        "-l",
        help="Maximim upload size in MB (set to 0 for no limit).",
        show_default=True,
        required=True,
        default=10,
        type=click.IntRange(min=0),
        cls=ResourceOption,
    )
    @functools.wraps(f)
    def wrapper_reusable_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_reusable_options
