# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""CHASE-DB1 dataset for Vessel Segmentation."""

import os
import pathlib

import PIL.Image
import pkg_resources
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.models.transforms import crop_image_to_mask
from mednet.libs.common.utils.rc import load_rc
from mednet.libs.segmentation.data.typing import (
    SegmentationRawDataLoader as _SegmentationRawDataLoader,
)
from torchvision import tv_tensors
from torchvision.transforms.functional import to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class SegmentationRawDataLoader(_SegmentationRawDataLoader):
    """A specialized raw-data-loader for the Chase-db1 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

        self._pkg_path = pathlib.Path(
            pkg_resources.resource_filename(__name__, "masks")
        )

    def sample(self, sample: tuple[str, str, str]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = to_tensor(PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB"))

        target = to_tensor(
            PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        )

        mask = to_tensor(
            PIL.Image.open(self._pkg_path / sample[2]).convert(mode="1", dither=None)
        )

        tensor = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Image(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(crop_image_to_mask(mask, mask))

        return tensor, dict(target=target, mask=mask, name=sample[0])  # type: ignore[arg-type]


class DataModule(CachingDataModule):
    """CHASE-DB1 dataset for Vessel Segmentation.

    The CHASE_DB1 is a retinal vessel reference dataset acquired from multiethnic
    school children. This database is a part of the Child Heart and Health Study in
    England (CHASE), a cardiovascular health survey in 200 primary schools in
    London, Birmingham, and Leicester. The ocular imaging was carried out in
    46 schools and demonstrated associations between retinal vessel tortuosity and
    early risk factors for cardiovascular disease in over 1000 British primary
    school children of different ethnic origin. The retinal images of both of the
    eyes of each child were recorded with a hand-held Nidek NM-200-D fundus camera.
    The images were captured at 30 degrees FOV camera. The dataset of images are
    characterized by having nonuniform back-ground illumination, poor contrast of
    blood vessels as compared with the background and wider arteriolars that have a
    bright strip running down the centre known as the central vessel reflex.

    * Reference: [CHASEDB1-2012]_
    * Original resolution (height x width): 960 x 999
    * Split reference: [CHASEDB1-2012]_
    * Protocol ``first-annotator``:

    * Training samples: 8 (including labels from annotator "1stHO")
    * Test samples: 20 (including labels from annotator "1stHO")

    * Protocol ``second-annotator``:

    * Training samples: 8 (including labels from annotator "2ndHO")
    * Test samples: 20 (including labels from annotator "2ndHO")

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=SegmentationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
