# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""HRF dataset for Vessel Segmentation."""

import os
import pathlib

import PIL.Image
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.models.transforms import crop_image_to_mask
from mednet.libs.common.utils.rc import load_rc
from mednet.libs.segmentation.data.typing import (
    SegmentationRawDataLoader as _SegmentationRawDataLoader,
)
from torchvision import tv_tensors
from torchvision.transforms.functional import to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class SegmentationRawDataLoader(_SegmentationRawDataLoader):
    """A specialized raw-data-loader for the drishtigs1hrf dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: tuple[str, str, str]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = to_tensor(PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB"))

        target = to_tensor(
            PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        )

        mask = to_tensor(
            PIL.Image.open(self.datadir / sample[2]).convert(mode="1", dither=None)
        )

        tensor = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Image(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(crop_image_to_mask(mask, mask))

        return tensor, dict(target=target, mask=mask, name=sample[0])  # type: ignore[arg-type]


class DataModule(CachingDataModule):
    """HRF dataset for Vessel Segmentation.

    The database includes 15 images of each healthy, diabetic retinopathy (DR), and
    glaucomatous eyes.  It contains a total  of 45 eye fundus images with a
    resolution of 3304 x 2336. One set of ground-truth vessel annotations is
    available.

    * Reference: [HRF-2013]_
    * Original resolution (height x width): 2336 x 3504
    * Configuration resolution: 1168 x 1648 (after specific cropping and rescaling)
    * Split reference: [ORLANDO-2017]_
    * Protocol ``default``:

    * Training samples: 15 (including labels)
    * Test samples: 30 (including labels)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=SegmentationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
