# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""RIM-ONE r3 (training set) for Cup Segmentation."""

import os
import pathlib

import PIL.Image
import pkg_resources
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import make_split
from mednet.libs.common.data.typing import Sample
from mednet.libs.common.utils.rc import load_rc
from mednet.libs.segmentation.data.typing import (
    SegmentationRawDataLoader as _SegmentationRawDataLoader,
)
from torchvision import tv_tensors
from torchvision.transforms.functional import crop, to_tensor

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class SegmentationRawDataLoader(_SegmentationRawDataLoader):
    """A specialized raw-data-loader for the rimoner3 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

        self._pkg_path = pathlib.Path(
            pkg_resources.resource_filename(__name__, "masks")
        )

    def crop_stereo_image(self, image):
        return crop(image, 0, 0, 1424, 1072)

    def sample(self, sample: tuple[str, str, str]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = to_tensor(
            self.crop_stereo_image(
                PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
            )
        )
        target = to_tensor(
            self.crop_stereo_image(
                PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
            )
        )
        mask = to_tensor(
            self.crop_stereo_image(
                PIL.Image.open(self._pkg_path / sample[2]).convert(
                    mode="1", dither=None
                )
            )
        )

        tensor = tv_tensors.Image(image)
        target = tv_tensors.Image(target)
        mask = tv_tensors.Mask(mask)

        return tensor, dict(target=target, mask=mask, name=sample[0])  # type: ignore[arg-type]


class DataModule(CachingDataModule):
    """RIM-ONE r3 (training set) for Cup Segmentation.

    The dataset contains 159 stereo eye fundus images with a resolution of 2144 x
    1424. The right part of the stereo image is disregarded. Two sets of
    ground-truths for optic disc and optic cup are available. The first set is
    commonly used for training and testing. The second set acts as a “human”
    baseline.  A third set, composed of annotation averages may also be used for
    training and evaluation purposes.

    * Reference: [RIMONER3-2015]_
    * Original resolution (height x width): 1424 x 1072
    * Split reference: [MANINIS-2016]_
    * Protocols ``optic-disc-exp1``, ``optic-cup-exp1``, ``optic-disc-exp2``,
      ``optic-cup-exp2``, ``optic-disc-avg`` and ``optic-cup-avg``
    * Training: 99
    * Test: 60

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=SegmentationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
        )
