# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Defines functionality for the evaluation of predictions."""

import logging
import pathlib
import typing

import h5py
import numpy
import numpy.typing
from tqdm import tqdm

logger = logging.getLogger(__name__)

SUPPORTED_METRIC_TYPE = typing.Literal[
    "precision", "recall", "specificity", "accuracy", "jaccard", "f1"
]
"""Supported metrics for evaluation of counts."""


def _tricky_division(n: float, d: float):
    """Divide n by d, or 0.0 in case of a division by zero.

    Parameters
    ----------
    n
        The number to divide.
    d
        The divisor.

    Returns
    -------
        Result of the division.
    """

    return n / (d + (d == 0))


def precision(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the precision given true/false positive/negative counts.

    P, AKA positive predictive value (PPV).  It corresponds arithmetically to
    ``tp/(tp+fp)``.  In the case ``tp+fp == 0``, this function returns zero for
    precision.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The precision.
    """
    return _tricky_division(tp, tp + fp)


def recall(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the recall given true/false positive/negative counts.

    R, AKA sensitivity, hit rate, or true positive rate (TPR).  It corresponds
    arithmetically to ``tp/(tp+fn)``.  In the special case where ``tp+fn ==
    0``, this function returns zero for recall.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The recall.
    """
    return _tricky_division(tp, tp + fn)


def specificity(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the specificity given true/false positive/negative counts.

    S, AKA selectivity or true negative rate (TNR).  It corresponds
    arithmetically to ``tn/(tn+fp)``.  In the special case where ``tn+fp ==
    0``, this function returns zero for specificity.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The specificity.
    """
    return _tricky_division(tn, fp + tn)


def accuracy(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the accuracy given true/false positive/negative counts.

    A, see `Accuracy
    <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_. is the
    proportion of correct predictions (both true positives and true negatives)
    among the total number of pixels examined.  It corresponds arithmetically
    to ``(tp+tn)/(tp+tn+fp+fn)``.  This measure includes both true-negatives
    and positives in the numerator, what makes it sensitive to data or regions
    without annotations.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The accuracy.
    """
    return _tricky_division(tp + tn, tp + fp + fn + tn)


def jaccard(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the Jaccard index given true/false positive/negative counts.

    J, see `Jaccard Index or Similarity
    <https://en.wikipedia.org/wiki/Jaccard_index>`_.  It corresponds
    arithmetically to ``tp/(tp+fp+fn)``.  In the special case where ``tn+fp+fn
    == 0``, this function returns zero for the Jaccard index. The Jaccard index
    depends on a TP-only numerator, similarly to the F1 score.  For regions
    where there are no annotations, the Jaccard index will always be zero,
    irrespective of the model output.  Accuracy may be a better proxy if one
    needs to consider the true abscence of annotations in a region as part of
    the measure.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The Jaccard index.
    """
    return _tricky_division(tp, tp + fp + fn)


def f1_score(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the F1 score given true/false positive/negative counts.

    F1, see `F1-score <https://en.wikipedia.org/wiki/F1_score>`_.  It
    corresponds arithmetically to ``2*P*R/(P+R)`` or ``2*tp/(2*tp+fp+fn)``. In
    the special case where ``P+R == (2*tp+fp+fn) == 0``, this function returns
    zero for the Jaccard index.  The F1 or Dice score depends on a TP-only
    numerator, similarly to the Jaccard index.  For regions where there are no
    annotations, the F1-score will always be zero, irrespective of the model
    output.  Accuracy may be a better proxy if one needs to consider the true
    abscence of annotations in a region as part of the measure.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The F1-score.
    """
    return _tricky_division(2 * tp, (2 * tp) + fp + fn)


_METRIC_MAPPING = {
    k: v
    for k, v in zip(
        typing.get_args(SUPPORTED_METRIC_TYPE),
        [precision, recall, specificity, accuracy, jaccard, f1_score],
    )
}
"""Maps supported metric names and callables implementing them."""


def name2metric(
    name: SUPPORTED_METRIC_TYPE,
) -> typing.Callable[[int, int, int, int], float]:
    """Convert a string name to a callable for summarizing counts.

    Parameters
    ----------
    name
        The name of the metric to be looked up.

    Returns
    -------
        A callable that summarizes counts into single floating-point number.
    """
    return _METRIC_MAPPING[name]


def all_metrics(tp: int, fp: int, tn: int, fn: int) -> list[float]:
    """Compute all available metrics at once.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        All supported metrics in the order defined by
        :py:obj:`SUPPORTED_METRIC_TYPE`.
    """
    return [k(tp, fp, tn, fn) for k in _METRIC_MAPPING.values()]


def tfpn_masks(
    pred: numpy.typing.NDArray[numpy.float32],
    gt: numpy.typing.NDArray[numpy.bool_],
    threshold: float,
) -> tuple[
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
]:
    """Calculate true and false positives and negatives.

    All input arrays should have matching sizes.

    Parameters
    ----------
    pred
        Pixel-wise predictions as output by your model.
    gt
        Ground-truth (annotations).
    threshold
        A particular threshold in which to calculate the performance
        measures.  Values at this threshold are counted as positives.

    Returns
    -------
    tp
        Boolean array with true positives, considering all observations.
    fp
        Boolean array with false positives, considering all observations.
    tn
        Boolean array with true negatives, considering all observations.
    fn
        Boolean array with false negatives, considering all observations.
    """

    binary_pred = pred >= threshold
    tp = numpy.logical_and(binary_pred, gt)
    fp = numpy.logical_and(binary_pred, numpy.logical_not(gt))
    tn = numpy.logical_and(numpy.logical_not(binary_pred), numpy.logical_not(gt))
    fn = numpy.logical_and(numpy.logical_not(binary_pred), gt)
    return tp, fp, tn, fn


def get_counts_for_threshold(
    pred: numpy.typing.NDArray[numpy.float32],
    gt: numpy.typing.NDArray[numpy.bool_],
    mask: numpy.typing.NDArray[numpy.bool_],
    threshold: float,
) -> tuple[int, int, int, int]:
    """Calculate counts on one single sample, for a specific threshold.

    Parameters
    ----------
    pred
        Array with pixel-wise predictions.
    gt
        Array with ground-truth (annotations).
    mask
        Array with region mask marking parts to ignore.
    threshold
        A particular threshold in which to calculate the performance
        measures.

    Returns
    -------
        The true positives, false positives, true negatives and false
        negatives, in that order.
    """

    tp, fp, tn, fn = tfpn_masks(pred, gt, threshold)
    tp = numpy.logical_and(tp, mask)
    fp = numpy.logical_and(fp, mask)
    tn = numpy.logical_and(tn, mask)
    fn = numpy.logical_and(fn, mask)
    return tp.sum(), fp.sum(), tn.sum(), fn.sum()


def load_count(
    prediction_path: pathlib.Path,
    predictions: typing.Sequence[str],
    thresholds: numpy.typing.NDArray[numpy.float64],
) -> numpy.typing.NDArray[numpy.uint64]:
    """Count true/false positive/negatives for the subset.

    This function will load predictions from their store location and will
    **cumulatively** count the number of true positives, false positives, true
    negatives and false negatives across the various ``thresholds``.  This
    alternative provides a memory-bound way to compute the performance of
    splits with potentially very large images or including a large/very large
    number of samples.  Unfortunately, sklearn does not provide functions to
    compute standard metrics from true/false positive/negative counts, which
    implies one needs to make use of further functions defined in this module
    to compute such metrics.  Alternatively, you may look into
    :py:func:`load_predictions`, if you want to use sklearn functions to
    compute metrics.

    Parameters
    ----------
    prediction_path
        Base directory where the prediction files (HDF5) were stored.
    predictions
        A list of relative sample prediction paths to consider for measurement.
    thresholds
        A sequence of thresholds to be applied on ``predictions``, when
        evaluating true/false positive/negative counts.

    Returns
    -------
        A 2-D array with shape ``(len(thresholds), 4)``, where each row
        contains to the counts of true positives, false positives, true
        negatives and false negatives, for the related threshold, and for the
        **whole dataset**.
    """
    data = numpy.zeros((len(thresholds), 4), dtype=numpy.uint64)
    for sample in tqdm(predictions, desc="sample"):
        with h5py.File(prediction_path / sample[1], "r") as f:
            pred = numpy.array(f.get("prediction"))  # float32
            gt = numpy.array(f.get("target"))  # boolean
            mask = numpy.array(f.get("mask"))  # boolean
        data += numpy.array(
            [get_counts_for_threshold(pred, gt, mask, k) for k in thresholds],
            dtype=numpy.uint64,
        )
    return data


def load_predictions(
    prediction_path: pathlib.Path,
    predictions: typing.Sequence[str],
) -> tuple[numpy.typing.NDArray[numpy.float32], numpy.typing.NDArray[numpy.bool_]]:
    """Load predictions and ground-truth from HDF5 files.

    Loading pixel-data as simple binary predictions with associated labels
    allows using sklearn library to compute most metrics defined in this
    module.  Note however that computing metrics this way requires
    pre-allocation of a potentially large vector, which depends on the number
    of samples and the size of said samples.  This may not work well for very
    large datasets of large/very large images.  Currently, the evaluation
    system uses :py:func:`load_count` instead, which loads and pre-computes the
    number of true/false positives/negatives using a list of candidate
    thresholds.

    Parameters
    ----------
    prediction_path
        Base directory where the prediction files (HDF5) were stored.
    predictions
        A list of relative sample prediction paths to consider for measurement.

    Returns
    -------
        Two 1-D arrays containing a linearized version of pixel predictions
        (probability) and matching ground-truth.
    """

    # peak prediction size and number of samples
    with h5py.File(prediction_path / predictions[0][1], "r") as f:
        elements = numpy.array(f.get("prediction").shape).prod()
    size = len(predictions) * elements
    logger.info(
        f"Data loading will require ({elements} x {len(predictions)} x 5 =) "
        f"{size*5/(1024*1024):.0f} MB of RAM"
    )

    # now load the data
    pred_array = numpy.empty((size,), dtype=numpy.float32)
    gt_array = numpy.empty((size,), dtype=numpy.bool_)
    for i, sample in enumerate(tqdm(predictions, desc="sample")):
        with h5py.File(prediction_path / sample[1], "r") as f:
            mask = numpy.array(f.get("mask"))  # boolean
            pred = numpy.array(f.get("prediction"))  # float32
            pred *= mask.astype(numpy.float32)
            gt = numpy.array(f.get("target"))  # boolean
            gt &= mask
            pred_array[i * elements : (i + 1) * elements] = pred.flatten()
            gt_array[i * elements : (i + 1) * elements] = gt.flatten()

    return pred_array, gt_array


def compute_metric(
    counts: numpy.typing.NDArray[numpy.uint64],
    metric: typing.Callable[[int, int, int, int], float]
    | typing.Callable[[int, int, int, int], tuple[float, ...]],
) -> numpy.typing.NDArray[numpy.float64]:
    """Compute ``metric`` for every row of ``counts``.

    Parameters
    ----------
    counts
        A 2-D array with shape ``(*, 4)``, where each row contains to the
        counts of true positives, false positives, true negatives and false
        negatives, that need to be evaluated.
    metric
        A callable that takes 4 integers representing true positives, false
        positives, true negatives and false negatives, and outputs one or more
        floating-point metrics.

    Returns
    -------
        An 1-D array containing the provided metric computed alongside the
        first dimension and as many columns as ```metric`` provides in each
        call.
    """
    return numpy.array([metric(*k) for k in counts], dtype=numpy.float64)


# def _compare_annotators_worker(
#     baseline_sample: tuple[str, str],
#     other_sample: tuple[str, str],
#     name: str,
#     output_folder: str | None,
#     overlayed_folder: str | None,
# ) -> tuple[str, pandas.DataFrame]:
#     """Run all of the comparison steps on a single sample pair.

#     Parameters
#     ----------
#     baseline_sample
#         Sample to be processed, containing the stem of the filepath
#         relative to the database root and a path the hdf5 file in which
#         the image, target and mask are stored.
#     other_sample
#         Another sample that is identical to the first, but has a different
#         mask (drawn by a different annotator).
#     name
#         The local name of the dataset (e.g. ``train``, or ``test``), to be
#         used when saving measures files.
#     output_folder
#         If not ``None``, then outputs a copy of the evaluation for this
#         sample in CSV format at this directory, but respecting the sample
#         ``stem``.
#     overlayed_folder
#         If not ``None``, then outputs a version of the input image with
#         predictions overlayed, in PNG format, but respecting the sample
#         ``stem``.

#     Returns
#     -------
#         The unique sample stem and a Dataframe containing the evaluation
#         performance on this sample.
#     """

#     assert baseline_sample[0] == other_sample[0], (
#         f"Mismatch between "
#         f"datasets for second-annotator analysis "
#         f"({baseline_sample[0]} != {other_sample[0]}).  This "
#         f"typically occurs when the second annotator (`other`) "
#         f"comes from a different dataset than the `baseline` dataset"
#     )

#     stem = baseline_sample[0]
#     image = baseline_sample[1]
#     gt = baseline_sample[2]
#     pred = other_sample[2]  # works as a prediction
#     mask = None if len(baseline_sample) < 4 else baseline_sample[3]
#     retval = _sample_measures(pred, gt, mask, 2)

#     if output_folder is not None:
#         fullpath = output_folder / "second-annotator" / name / f"{stem}.csv"
#         tqdm.write(f"Saving {fullpath}...")
#         fullpath.parent.mkdir(parents=True, exist_ok=True)
#         retval.to_csv(fullpath)

#     if overlayed_folder is not None:
#         overlay_image = _sample_analysis(
#             image, pred, gt, mask, threshold=0.5, overlay=True
#         )
#         fullpath = overlayed_folder / "second-annotator" / name / f"{stem}.png"
#         tqdm.write(f"Saving {fullpath}...")
#         fullpath.parent.mkdir(parents=True, exist_ok=True)
#         overlay_image.save(fullpath)

#     return stem, retval


# def compare_annotators(
#     baseline: torch.utils.data.Dataset,
#     other: torch.utils.data.Dataset,
#     name: str,
#     output_folder: str,
#     overlayed_folder: str | None | None = None,
#     parallel: str | None = -1,
# ):
#     """Compare annotations on the **same** dataset.

#     Parameters
#     ----------
#     baseline
#         A dataset to iterate on, containing the baseline annotations.
#     other
#         A second dataset, with the same samples as ``baseline``, but annotated
#         by a different annotator than in the first dataset.  The key values
#         must much between ``baseline`` and this dataset.
#     name
#         The local name of this dataset (e.g. ``train-second-annotator``, or
#         ``test-second-annotator``), to be used when saving measures files.
#     output_folder
#         Folder where to store results.
#     overlayed_folder
#         If not ``None``, then it should be the name of a folder where to store
#         overlayed versions of the images and ground-truths.
#     parallel
#         If set to a value different >= 0, uses multiprocessing for estimating
#         thresholds for each sample through a processing pool.  A value of zero
#         will create as many processes in the pool as cores in the machine.  A
#         negative value disables multiprocessing altogether.  A value greater
#         than zero will spawn as many processes as requested.
#     """

#     logger.info(f"Output folder: {output_folder}")
#     output_folder.mkdir(parents=True, exist_ok=True)

#     # Collect overall measures
#     data = {}

#     if parallel < 0:  # turns off multiprocessing
#         for baseline_sample, other_sample in tqdm(
#             list(zip(baseline, other)),
#             desc="samples",
#             leave=False,
#             disable=None,
#         ):
#             k, v = _compare_annotators_worker(
#                 (
#                     baseline_sample,
#                     other_sample,
#                     name,
#                     output_folder,
#                     overlayed_folder,
#                 )
#             )
#             data[k] = v
#     else:
#         parallel = parallel or multiprocessing.cpu_count()
#         with (
#             multiprocessing.Pool(processes=parallel) as pool,
#             tqdm(
#                 total=len(baseline),
#                 desc="sample",
#             ) as pbar,
#         ):
#             for k, v in pool.imap_unordered(
#                 _compare_annotators_worker,
#                 zip(
#                     baseline,
#                     other,
#                     itertools.repeat(name),
#                     itertools.repeat(output_folder),
#                     itertools.repeat(overlayed_folder),
#                 ),
#             ):
#                 pbar.update()
#                 data[k] = v

#     measures = _summarize(data)
#     measures.drop(0, inplace=True)  # removes threshold == 0.0, keeps 0.5 only

#     measures_path = output_folder / "second-annotator" / f"{name}.csv"

#     measures_path.parent.mkdir(parents=True, exist_ok=True)
#     logger.info(f"Saving summaries over all input images at {measures_path}...")
#     measures.to_csv(measures_path)

#     maxf1 = measures["mean_f1_score"].max()
#     logger.info(f"F1-score of {maxf1:.5f} (second annotator; threshold=0.5)")
