# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Loss implementations."""

import torch


class WeightedBCELogitsLoss(torch.nn.Module):
    """Calculates sum of weighted cross entropy loss.

    Implements Equation 1 in [MANINIS-2016]_.  The weight depends on the
    current proportion between negatives and positives in the ground-
    truth sample being analyzed.
    """

    def __init__(self):
        super().__init__()

    def forward(
        self, sample: torch.Tensor, target: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        sample
            Value produced by the model to be evaluated, with the shape ``[n, c,
            h, w]``.
        target
            Ground-truth information with the shape ``[n, c, h, w]``.
        mask
            Mask to be use for specifying the region of interest where to
            compute the loss, with the shape ``[n, c, h, w]``.

        Returns
        -------
            The average loss for all input data.
        """

        # calculates the proportion of negatives to the total number of pixels
        # available in the masked region
        valid = mask > 0.5
        num_pos = target[valid].sum()
        num_neg = valid.sum() - num_pos
        pos_weight = num_neg / num_pos
        return torch.nn.functional.binary_cross_entropy_with_logits(
            sample[valid],
            target[valid],
            reduction="mean",
            pos_weight=pos_weight,
        )


class SoftJaccardBCELogitsLoss(torch.nn.Module):
    r"""Implement the generalized loss function of Equation (3) in.

    [IGLOVIKOV-2018]_, with J being the Jaccard distance, and H, the Binary
    Cross-Entropy Loss:

    .. math::

       L = \alpha H + (1-\alpha)(1-J)

    Our implementation is based on :py:class:`torch.nn.BCEWithLogitsLoss`.

    Parameters
    ----------
    alpha
        Determines the weighting of J and H. Default: ``0.7``.
    """

    def __init__(self, alpha: float = 0.7):
        super().__init__()
        self.alpha = alpha

    def forward(
        self, tensor: torch.Tensor, target: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        tensor
            Value produced by the model to be evaluated, with the shape ``[n, c,
            h, w]``.
        target
            Ground-truth information with the shape ``[n, c, h, w]``.
        mask
            Mask to be use for specifying the region of interest where to
            compute the loss, with the shape ``[n, c, h, w]``.

        Returns
        -------
            Loss, in a single entry.
        """

        eps = 1e-8
        valid = mask > 0.5
        probabilities = torch.sigmoid(tensor[valid])
        intersection = (probabilities * target[valid]).sum()
        sums = probabilities.sum() + target[valid].sum()
        j = intersection / (sums - intersection + eps)

        # this implements the support for looking just into the RoI
        h = torch.nn.functional.binary_cross_entropy_with_logits(
            tensor[valid], target[valid], reduction="mean"
        )
        return (self.alpha * h) + ((1 - self.alpha) * (1 - j))


class MultiWeightedBCELogitsLoss(WeightedBCELogitsLoss):
    """Weighted Binary Cross-Entropy Loss for multi-layered inputs (e.g. for
    Holistically-Nested Edge Detection in [XIE-2015]_).
    """

    def __init__(self):
        super().__init__()

    def forward(
        self, tensor: torch.Tensor, target: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        tensor
            Value produced by the model to be evaluated, with the shape ``[L,
            n, c, h, w]``.
        target
            Ground-truth information with the shape ``[n, c, h, w]``.
        mask
            Mask to be use for specifying the region of interest where to
            compute the loss, with the shape ``[n, c, h, w]``.

        Returns
        -------
            The average loss for all input data.
        """

        return torch.cat(
            [
                super(MultiWeightedBCELogitsLoss, self)
                .forward(i, target, mask)
                .unsqueeze(0)
                for i in tensor
            ]
        ).mean()


class MultiSoftJaccardBCELogitsLoss(SoftJaccardBCELogitsLoss):
    """Implements  Equation 3 in [IGLOVIKOV-2018]_ for the multi-output
    networks such as HED or Little W-Net.

    Parameters
    ----------
    alpha : float
        Determines the weighting of SoftJaccard and BCE. Default: ``0.3``.
    """

    def __init__(self, alpha: float = 0.7):
        super().__init__(alpha=alpha)

    def forward(
        self, tensor: torch.Tensor, target: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        tensor
            Value produced by the model to be evaluated, with the shape ``[L,
            n, c, h, w]``.
        target
            Ground-truth information with the shape ``[n, c, h, w]``.
        mask
            Mask to be use for specifying the region of interest where to
            compute the loss, with the shape ``[n, c, h, w]``.

        Returns
        -------
            The average loss for all input data.
        """

        return torch.cat(
            [
                super(MultiSoftJaccardBCELogitsLoss, self)
                .forward(i, target, mask)
                .unsqueeze(0)
                for i in tensor
            ]
        ).mean()


# class MixJacLoss(torch.nn.Module):
#     """Implements Mix Jaccard Loss.

#     Parameters
#     ----------
#     lambda_u
#         Determines the weighting of SoftJaccard and BCE.
#     jacalpha
#         Determines the weighting of J and H.
#     size_average
#         By default, the losses are averaged over each loss element in the
#         batch. Note that for some losses, there are multiple elements per
#         sample. If the field `size_average` is set to ``False``, the losses
#         are instead summed for each minibatch. Ignored when `reduce` is
#         ``False``. Default: ``True``.
#     reduce
#         By default, the losses are averaged or summed over observations for
#         each minibatch depending on `size_average`. When `reduce` is
#         ``False``, returns a loss per batch element instead and ignores
#         `size_average`. Default: ``True``.
#     reduction
#         Specifies the reduction to apply to the output: ``'none'`` |
#         ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
#         ``'mean'``: the sum of the output will be divided by the number of
#         elements in the output, ``'sum'``: the output will be summed. Note:
#         `size_average` and `reduce` are in the process of being deprecated,
#         and in the meantime, specifying either of those two args will
#         override `reduction`. Default: ``'mean'``.
#     """

#     def __init__(
#         self,
#         lambda_u: int = 100,
#         jacalpha=0.7,
#         size_average=None,
#         reduce=None,
#         reduction="mean",
#     ):
#         super().__init__(size_average, reduce, reduction)
#         self.lambda_u = lambda_u
#         self.labeled_loss = SoftJaccardBCELogitsLoss(alpha=jacalpha)
#         self.unlabeled_loss = torch.nn.BCEWithLogitsLoss()

#     def forward(
#         self,
#         tensor: torch.Tensor,
#         target: torch.Tensor,
#         unlabeled_tensor: torch.Tensor,
#         unlabeled_target: torch.Tensor,
#         ramp_up_factor: float,
#     ) -> tuple:
#         """Forward pass.

#         Parameters
#         ----------
#         tensor
#             Value produced by the model to be evaluated, with the shape ``[L,
#             n, c, h, w]``.
#         target
#             Ground-truth information with the shape ``[n, c, h, w]``.

#         unlabeled_tensor

#         unlabeled_target

#         ramp_up_factor

#         Returns
#         -------
#         list
#         """
#         ll = self.labeled_loss(tensor, target)
#         ul = self.unlabeled_loss(unlabeled_tensor, unlabeled_target)

#         loss = ll + self.lambda_u * ramp_up_factor * ul
#         return loss, ll, ul
