# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Contains the inverse :py:func:`torch.utils.data.default_collate`."""

import typing

from mednet.libs.common.data.typing import Sample

from .typing import SegmentationPrediction


def _as_predictions(
    samples: typing.Iterable[Sample],
) -> list[SegmentationPrediction]:
    """Take a list of separated batch predictions and transforms it into a list
    of formal predictions.

    Parameters
    ----------
    samples
        A sequence of samples as returned by :py:func:`separate`.

    Returns
    -------
        A list of typed predictions that can be saved to disk.
    """
    return [(v[1]["name"], v[1]["target"], v[1]["mask"], v[0]) for v in samples]


def separate(batch: Sample) -> list[SegmentationPrediction]:
    """Separate a collated batch, reconstituting its samples.

    This function implements the inverse of
    :py:func:`torch.utils.data.default_collate`, and can separate, into
    samples, batches of data with different attributes.  It follows the inverse
    path of that function, and implements the following separation algorithms:

    * :class:`torch.Tensor` -> :class:`torch.Tensor` (with a removed outer
      dimension, via :py:func:`torch.flatten`)
    * ``typing.Mapping[K, V[]]`` -> ``[dict[K, V_1], dict[K, V_2], ...]``

    Parameters
    ----------
    batch
        A batch, as output by torch model forwarding.

    Returns
    -------
        A list of predictions that contains the predictions and associated metadata
        for each processed sample.
    """

    # as of now, this is really simple - to be made more complex upon need.
    metadata = [
        {key: value[i] for key, value in batch[1].items()} for i in range(len(batch[0]))
    ]

    return _as_predictions(zip(batch[0], metadata))
