# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import AliasedGroup, verbosity_option
from clapper.logging import setup
from mednet.libs.common.scripts.config import copy as copy_
from mednet.libs.common.scripts.config import describe as describe_
from mednet.libs.common.scripts.config import list_ as list__

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.group(cls=AliasedGroup)
def config():
    """Command for listing, describing and copying configuration resources."""
    pass


@config.command(
    name="list",
    epilog="""Examples:

\b
  1. Lists all configuration resources (type: mednet.libs.segmentation.config) installed:

     .. code:: sh

        mednet segmentation config list


\b
  2. Lists all configuration resources and their descriptions (notice this may
     be slow as it needs to load all modules once):

     .. code:: sh

        mednet segmentation config list -v

""",
)
@verbosity_option(logger=logger)
def list_(verbose) -> None:  # numpydoc ignore=PR01
    """List configuration files installed."""
    list__("mednet.libs.segmentation.config", verbose)


@config.command(
    epilog="""Examples:

\b
  1. Describe the Drive dataset configuration:

     .. code:: sh

        mednet segmentation config describe drive


\b
  2. Describe the Drive dataset configuration and lists its
     contents:

     .. code:: sh

        mednet segmentation config describe deive -v

""",
)
@click.argument(
    "name",
    required=True,
    nargs=-1,
)
@verbosity_option(logger=logger)
def describe(name, verbose) -> None:  # numpydoc ignore=PR01
    """Describe a specific configuration file."""
    describe_(name, "mednet.libs.segmentation.config", verbose)


@config.command(
    epilog="""Examples:

\b
  1. Make a copy of one of the stock configuration files locally, so it can be
     adapted:

     .. code:: sh

        $ mednet segmentation config copy drive -vvv newdataset.py

""",
)
@click.argument(
    "source",
    required=True,
    nargs=1,
)
@click.argument(
    "destination",
    required=True,
    nargs=1,
)
@verbosity_option(logger=logger, expose_value=False)
def copy(source, destination) -> None:  # numpydoc ignore=PR01
    """Copy a specific configuration resource so it can be modified locally."""
    copy_(source, destination, "mednet.libs.segmentation.config")
