# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import AliasedGroup, verbosity_option
from clapper.logging import setup
from mednet.libs.common.scripts.database import check as check_
from mednet.libs.common.scripts.database import list_ as list__

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def _get_raw_databases() -> dict[str, dict[str, str]]:
    """Return a list of all supported (raw) databases.

    Returns
    -------
    dict[str, dict[str, str]]
        Dictionary where keys are database names, and values are dictionaries
        containing two string keys:

        * ``module``: the full Pythonic module name (e.g.
        ``mednet.libs.segmentation.data.drive``).
        * ``datadir``: points to the user-configured data directory for the
        current dataset, if set, or ``None`` otherwise.
    """

    import importlib
    import pkgutil

    from mednet.libs.common.utils.rc import load_rc

    from ..config import data

    user_configuration = load_rc()

    retval = {}
    for k in pkgutil.iter_modules(data.__path__):
        for j in pkgutil.iter_modules(
            [next(iter(data.__path__)) + f"/{k.name}"],
        ):
            if j.name == "datamodule":
                # this is a submodule that can read raw data files
                module = importlib.import_module(
                    f".{j.name}",
                    data.__package__ + f".{k.name}",
                )
                if hasattr(module, "CONFIGURATION_KEY_DATADIR"):
                    retval[k.name] = dict(
                        module=module.__name__.rsplit(".", 1)[0],
                        datadir=user_configuration.get(
                            module.CONFIGURATION_KEY_DATADIR,
                        ),
                    )
                else:
                    retval[k.name] = dict(module=module.__name__)

    return retval


@click.group(cls=AliasedGroup)
def database() -> None:
    """Command for listing and verifying databases installed."""
    pass


@database.command(
    name="list",
    epilog="""Examples:

\b
    1. To install a database, set up its data directory ("datadir").  For
       example, to setup access to Drive files you downloaded locally at
       the directory "/path/to/drive/files", edit the RC file (typically
       ``$HOME/.config/mednet.toml``), and add a line like the following:

       .. code:: toml

          [datadir]
          drive = "/path/to/drive/files"

       .. note::

          This setting **is** case-sensitive.

\b
    2. List all raw databases supported (and configured):

       .. code:: sh

          $ mednet segmentation database list

""",
)
@verbosity_option(logger=logger, expose_value=False)
def list_():
    """List all supported and configured databases."""
    list__(_get_raw_databases())


@database.command(
    epilog="""Examples:

    1. Check if all files from the fold 'default' of the Drive
       database can be loaded:

       .. code:: sh

          mednet segmentation database check -vv drive

""",
)
@click.argument(
    "fold",
    nargs=1,
)
@click.option(
    "--limit",
    "-l",
    help="Limit check to the first N samples in each split dataset, making the "
    "check sensibly faster.  Set it to zero (default) to check everything.",
    required=True,
    type=click.IntRange(0),
    default=0,
)
@verbosity_option(logger=logger, expose_value=False)
def check(fold, limit):  # numpydoc ignore=PR01
    """Check file access on one or more DataModules."""
    check_("mednet.libs.segmentation.config", fold, limit)
