# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import pathlib
import typing

import click
from clapper.click import ResourceOption, verbosity_option
from clapper.logging import setup
from mednet.libs.common.scripts.click import ConfigCommand
from mednet.libs.segmentation.engine.evaluator import SUPPORTED_METRIC_TYPE

logger = setup("mednet")


def validate_threshold(threshold: float | str, splits: list[str]):
    """Validate the user threshold selection and returns parsed threshold.

    Parameters
    ----------
    threshold
        The threshold to validate.
    splits
        List of available splits.

    Returns
    -------
        The validated threshold.
    """
    if threshold is None:
        return 0.5

    try:
        # we try to convert it to float first
        threshold = float(threshold)
        if threshold < 0.0 or threshold > 1.0:
            raise ValueError("Float thresholds must be within range [0.0, 1.0]")
    except ValueError:
        if threshold not in splits:
            raise ValueError(
                f"Text thresholds should match dataset names, "
                f"but {threshold} is not available among the datasets provided ("
                f"({', '.join(splits)})"
            )

    return threshold


@click.command(
    entry_point_group="mednet.libs.segmentation.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
  1. Runs evaluation on an existing dataset configuration:

     .. code:: sh

        $ mednet segmentation evaluate -vv --predictions=path/to/predictions.json --output-folder=path/to/results
""",
)
@click.option(
    "--predictions",
    "-p",
    help="""Path to the JSON file describing available predictions. The actual
    predictions are supposed to lie on the same folder.""",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=False,
        path_type=pathlib.Path,
    ),
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    cls=ResourceOption,
)
# @click.option(
#     "--second-annotator",
#     "-a",
#     help="""A datamodule containing annotations from another annotator, that
#     will be compared to the ground-truth (reference annotator) in each
#     sample.""",
#     required=False,
#     default=None,
#     cls=ResourceOption,
#     show_default=True,
# )
@click.option(
    "--threshold",
    "-t",
    help="""This number is used to define positives and negatives from
    probability maps, and used to report metrics based on a threshold chosen *a
    priori*. It can be set to a floating-point value, or to the name of dataset
    split in ``--predictions``.
    """,
    default="0.5",
    show_default=True,
    required=False,
    cls=ResourceOption,
)
@click.option(
    "--metric",
    "-m",
    help="""If threshold is set to the name of a split in ``--predictions``,
    then this parameter defines the metric function to be used to evaluate the
    threshold at which the metric reaches its maximum value. All other splits
    are evaluated with respect to this threshold.""",
    default="f1",
    type=click.Choice(typing.get_args(SUPPORTED_METRIC_TYPE), case_sensitive=True),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--steps",
    "-s",
    help="""Number of steps for evaluating metrics on various splits. This
    value is used when drawing precision-recall plots, or when deciding the
    highest metric value on splits.""",
    default=100,
    type=click.IntRange(10),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def evaluate(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    metric: str,
    # second_annotator,
    steps: int,
    **_,  # ignored
):  # numpydoc ignore=PR01
    """Evaluate predictions (from a model) on a segmentation task."""

    import credible.curves
    import credible.plot
    import matplotlib.backends.backend_pdf
    import numpy
    import tabulate
    from mednet.libs.classification.engine.evaluator import NumpyJSONEncoder
    from mednet.libs.common.scripts.utils import (
        execution_metadata,
        save_json_with_backup,
    )
    from mednet.libs.segmentation.engine.evaluator import (
        all_metrics,
        compute_metric,
        load_count,
        name2metric,
        precision,
        recall,
        specificity,
    )

    evaluation_filename = "evaluation.json"
    evaluation_file = pathlib.Path(output_folder) / evaluation_filename

    with predictions.open("r") as f:
        predict_data = json.load(f)

    # register metadata
    json_data: dict[str, typing.Any] = execution_metadata()
    json_data.update(
        dict(
            predictions=str(predictions),
            output_folder=str(output_folder),
            threshold=threshold,
            metric=metric,
            steps=steps,
        ),
    )
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(evaluation_file.with_suffix(".meta.json"), json_data)

    threshold = validate_threshold(threshold, predict_data)
    threshold_list = numpy.arange(
        0.0, (1.0 + 1 / steps), 1 / steps, dtype=numpy.float64
    )

    # Compute counts for various splits
    eval_json_data: dict[str, dict[str, typing.Any]] = {}
    for split_name, samples in predict_data.items():
        logger.info(
            f"Counting true/false positive/negatives at split `{split_name}`..."
        )
        eval_json_data.setdefault(split_name, {})["counts"] = {
            k: v
            for k, v in zip(
                threshold_list, load_count(predictions.parent, samples, threshold_list)
            )
        }
        eval_json_data[split_name]["threshold_a_posteriori"] = True

    if isinstance(threshold, str):
        # Compute threshold on specified split, if required
        logger.info(f"Evaluating threshold on `{threshold}` split using " f"`{metric}`")
        metric_list = compute_metric(
            eval_json_data[threshold]["counts"].values(),
            name2metric(typing.cast(SUPPORTED_METRIC_TYPE, metric)),
        )
        threshold_index = metric_list.argmax()
        logger.info(f"Set --threshold={threshold_list[threshold_index]:.4f}")

        # Reset list of how thresholds are calculated on the recorded split
        for split_name in predict_data.keys():
            if split_name == threshold:
                continue
            eval_json_data[split_name]["threshold_a_posteriori"] = False

    else:
        # must figure out the closest threshold from the list we are using
        threshold_index = (numpy.abs(threshold_list - threshold)).argmin()
        logger.info(f"Set --threshold={threshold_list[threshold_index]:.4f}")

    logger.info("Tabulating performance summary...")
    table_format = "rst"
    output_table = output_folder / "evaluation.rst"
    metrics_available = list(typing.get_args(SUPPORTED_METRIC_TYPE))
    table_headers = ["Dataset", "threshold"] + metrics_available + ["auroc", "avgprec"]

    table_data = []
    for split_name in predict_data.keys():
        counts = list(eval_json_data[split_name]["counts"].values())
        base_metrics = all_metrics(*counts[threshold_index])
        table_data.append([split_name, threshold_list[threshold_index]] + base_metrics)
        eval_json_data[split_name].update(
            {k: v for k, v in zip(metrics_available, base_metrics)}
        )
        fpr_curve = 1.0 - numpy.array([specificity(*k) for k in counts])
        recall_curve = tpr_curve = numpy.array([recall(*k) for k in counts])
        precision_curve = numpy.array([precision(*k) for k in counts])
        table_data[-1] += [
            credible.curves.area_under_the_curve((fpr_curve, tpr_curve)),  # auc-roc
            credible.curves.average_metric(
                (precision_curve, recall_curve)
            ),  # average precision
        ]
        eval_json_data[split_name]["auc_score"] = table_data[-1][-2]
        eval_json_data[split_name]["average_precision_score"] = table_data[-1][-1]
        eval_json_data[split_name]["curves"] = dict(
            roc=dict(fpr=fpr_curve, tpr=tpr_curve, thresholds=threshold_list),
            precision_recall=dict(
                precision=precision_curve,
                recall=recall_curve,
                thresholds=threshold_list,
            ),
        )

    # records full result analysis to a JSON file
    evaluation_file = output_folder / "evaluation.json"
    logger.info(f"Saving evaluation results at `{evaluation_file}`...")
    with evaluation_file.open("w") as f:
        json.dump(eval_json_data, f, indent=2, cls=NumpyJSONEncoder)

    table = tabulate.tabulate(
        table_data,
        table_headers,
        tablefmt=table_format,
        floatfmt=".3f",
        stralign="right",
    )
    click.echo(table)
    logger.info(f"Saving table at {output_table}...")
    output_table.parent.mkdir(parents=True, exist_ok=True)
    with output_table.open("w") as f:
        f.write(table)

    logger.info("Plotting performance curves...")
    output_figure = output_folder / "evaluation.pdf"
    logger.info(f"Saving figures at {output_figure}...")
    with matplotlib.backends.backend_pdf.PdfPages(output_figure) as pdf:
        with credible.plot.tight_layout(
            ("False Positive Rate", "True Positive Rate"), "ROC"
        ) as (
            fig,
            ax,
        ):
            for split_name, data in eval_json_data.items():
                ax.plot(
                    data["curves"]["roc"]["fpr"],
                    data["curves"]["roc"]["tpr"],
                    label=f"{split_name} (AUC: {data['auc_score']:.2f})",
                )
                ax.legend(loc="best", fancybox=True, framealpha=0.7)
            pdf.savefig(fig)

        with credible.plot.tight_layout_f1iso(
            ("Recall", "Precision"), "Precison-Recall"
        ) as (
            fig,
            ax,
        ):
            for split_name, data in eval_json_data.items():
                ax.plot(
                    data["curves"]["precision_recall"]["precision"],
                    data["curves"]["precision_recall"]["recall"],
                    label=f"{split_name} (AP: {data['average_precision_score']:.2f})",
                )
                ax.legend(loc="best", fancybox=True, framealpha=0.7)
            pdf.savefig(fig)
