# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import clapper.click
import clapper.logging
import click
import mednet.libs.common.scripts.click
import mednet.libs.common.scripts.predict

logger = clapper.logging.setup(
    __name__.split(".")[0], format="%(levelname)s: %(message)s"
)


@click.command(
    entry_point_group="mednet.libs.segmentation.config",
    cls=mednet.libs.common.scripts.click.ConfigCommand,
    epilog="""Examples:

1. Run prediction on an existing DataModule configuration:

   .. code:: sh

      mednet segmentation predict -vv lwnet drive --weight=path/to/model.ckpt --output-folder=path/to/predictions

2. Enable multi-processing data loading with 6 processes:

   .. code:: sh

      mednet segmentation predict -vv lwnet drive --parallel=6 --weight=path/to/model.ckpt --output-folder=path/to/predictions

""",
)
@mednet.libs.common.scripts.predict.reusable_options
@clapper.click.verbosity_option(
    logger=logger, cls=clapper.click.ResourceOption, expose_value=False
)
def predict(
    output_folder,
    model,
    datamodule,
    batch_size,
    device,
    weight,
    parallel,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Run inference (generates scores) on all input images, using a pre-trained model."""

    from mednet.libs.common.engine.device import DeviceManager
    from mednet.libs.common.scripts.predict import (
        load_checkpoint,
        save_json_data,
        setup_datamodule,
    )
    from mednet.libs.common.scripts.utils import save_json_with_backup
    from mednet.libs.segmentation.engine.predictor import run

    predictions_meta_file = output_folder / "predictions.meta.json"

    setup_datamodule(datamodule, model, batch_size, parallel)
    model = load_checkpoint(model, weight)
    device_manager = DeviceManager(device)
    save_json_data(datamodule, model, device_manager, predictions_meta_file)

    json_predictions = run(model, datamodule, device_manager, output_folder)

    save_json_with_backup(output_folder / "predictions.json", json_predictions)
